// ChildView.cpp : implementation of the CChildView class
//

#include "stdafx.h"
#include "MVSTracker.h"
#include "ChildView.h"
#include "modplay.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

const int ROWHEIGHT = 17;
const int YOFFSET = ROWHEIGHT;
const int XOFFSET = 40;
const int CHANNEL_CX = 150;

const char * NoteScale[14] = 
{ 
   "C -", "C#", "D -", "D#", "E -", "F -", "F#", "G -", "G#", "A -", "A#", "B -", "- -", "END"
};

static unsigned char SongBuffer[65536];

const int CurOffsets[MAXITEMS * 2] =
{
   0,30,  // Note
   40,15, // Octave
   65,25, // Voice
   95,25,  // Effect
   125,25  // Value
};


/////////////////////////////////////////////////////////////////////////////
// CChildView

CChildView::CChildView()
{
   mFirstRow = 0;
   mcrMain = RGB(220,245,220);
   mcrMain5 = RGB(200,245,200);
   mcrHeader = RGB(180,245,180);
   mCurRow = 0;
   mCurX = 0;
   
   // default header
   mhdrPatternNum = 0;
   mhdrEditMode = true;
   mpCurPattern = &Global::Patterns[0];

   // so we don't have to reload the instrument
   m_LastInstrPlayed = 99;
}

CChildView::~CChildView()
{
}


BEGIN_MESSAGE_MAP(CChildView,CWnd )
	//{{AFX_MSG_MAP(CChildView)
	ON_WM_PAINT()
	ON_WM_VSCROLL()
	ON_WM_KEYDOWN()
	ON_WM_LBUTTONDOWN()
	ON_WM_ERASEBKGND()
	ON_WM_CHAR()
	ON_COMMAND(ID_EDIT_DECREASETRACKOCTAVE, OnEditDecreasetrackoctave)
	ON_COMMAND(ID_EDIT_INCREASETRACKOCTAVE, OnEditIncreasetrackoctave)
	ON_COMMAND(ID_PLAY_CURRENTPATTERN, OnPlayCurrentpattern)
	ON_COMMAND(ID_PLAY_SONG, OnPlaySong)
	ON_COMMAND(ID_FILE_EXPORTTESTROM, OnFileExporttestrom)
	ON_COMMAND(ID_FILE_SAVE, OnFileSave)
	ON_COMMAND(ID_FILE_OPEN, OnFileOpen)
	ON_COMMAND(ID_FILE_NEWPROJECT, OnFileNewproject)
	ON_COMMAND(ID_FILE_IMPORTMOD, OnFileImportmod)
	ON_COMMAND(ID_PLAY_STOP, OnPlayStop)
	ON_COMMAND(ID_FILE_EXPORT_MVS0,OnFileExportMVS0)
	ON_COMMAND(ID_FILE_EXPORTMDTESTROM,OnFileExportMdTestRom)

	ON_BN_CLICKED(IDC_PH_BUT_PLAY_ALL, OnPhButPlayAll)
	ON_BN_CLICKED(IDC_PH_BUT_PLAY, OnPhButPlay)
	ON_BN_CLICKED(IDC_PH_BUT_STOP, OnPhButStop)
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()


/////////////////////////////////////////////////////////////////////////////
// CChildView message handlers

BOOL CChildView::PreCreateWindow(CREATESTRUCT& cs) 
{
	if (!CWnd::PreCreateWindow(cs))
		return FALSE;

	cs.dwExStyle |= WS_EX_CLIENTEDGE;
	cs.style &= ~WS_BORDER;
	cs.lpszClass = AfxRegisterWndClass(CS_HREDRAW|CS_VREDRAW|CS_DBLCLKS, 
		::LoadCursor(NULL, IDC_ARROW), HBRUSH(COLOR_WINDOW+1), NULL);

	return TRUE;
}

void CChildView::OnPaint() 
{
	CPaintDC dc(this); // device context for painting

   int PatNum = Global::pMainFrm->ToolBar_GetPattern();
   mpCurPattern = &Global::Patterns[PatNum];

   PatEd_DrawHeader(&dc);
   PatEd_DrawPattern(&dc);
   PatEd_DrawCursor();
}



void CChildView::PatEd_DrawCursor()
{
   CDC *devc = GetDC();

   // only draw cursor if on screen
   CRect rClient; GetClientRect(&rClient);
   int NumRows = (rClient.bottom - rClient.top - YOFFSET) / ROWHEIGHT;

   if (mCurRow < mFirstRow || mCurRow > mFirstRow + NumRows - 1)
      return;

   // calc cursor position
   CRect rCursor;

   int Channel = mCurX / MAXITEMS;
   int Item = mCurX % MAXITEMS;

   rCursor.left = Channel * CHANNEL_CX + CurOffsets[Item*2] + XOFFSET;
   rCursor.top = rClient.top + YOFFSET + (ROWHEIGHT * (mCurRow - mFirstRow));
   rCursor.bottom = rCursor.top + ROWHEIGHT;
   rCursor.right = rCursor.left + CurOffsets[Item*2+1];

   devc->InvertRect(&rCursor);
}


void CChildView::PatEd_DrawHeader(CDC *devc)
{
   // fill background colour
   CRect rClient; GetClientRect(&rClient);
   rClient.bottom = YOFFSET;
   CBrush brh1(mcrHeader);
   devc->FillRect(&rClient, &brh1);

   devc->SetTextColor(0);
   devc->SetBkColor(mcrHeader);

   CBitmap bmTrack1;
   CBitmap bmTrack2;
   CBitmap bmTrack3;
   CBitmap bmTrack4;
   CDC dcMemory;
   dcMemory.CreateCompatibleDC(devc);


   if (Global::bChannelActive[0])
      bmTrack1.LoadBitmap(IDB_TRACK1_ON);
   else
      bmTrack1.LoadBitmap(IDB_TRACK1_OFF);
   CBitmap* pOldBitmap = dcMemory.SelectObject(&bmTrack1);
   devc->BitBlt(XOFFSET, 0, CHANNEL_CX, ROWHEIGHT, &dcMemory, 
      0, 0, SRCCOPY);

   if (Global::bChannelActive[1])
      bmTrack2.LoadBitmap(IDB_TRACK2_ON);
   else
      bmTrack2.LoadBitmap(IDB_TRACK2_OFF);
   dcMemory.SelectObject(&bmTrack2);
   devc->BitBlt(XOFFSET+CHANNEL_CX, 0, CHANNEL_CX, ROWHEIGHT, &dcMemory, 
      0, 0, SRCCOPY);

   if (Global::bChannelActive[2])
      bmTrack3.LoadBitmap(IDB_TRACK3_ON);
   else
      bmTrack3.LoadBitmap(IDB_TRACK3_OFF);
   dcMemory.SelectObject(&bmTrack3);
   devc->BitBlt(XOFFSET+(CHANNEL_CX*2), 0, CHANNEL_CX, ROWHEIGHT, &dcMemory, 
      0, 0, SRCCOPY);

   if (Global::bChannelActive[3])
      bmTrack4.LoadBitmap(IDB_TRACK4_ON);
   else
      bmTrack4.LoadBitmap(IDB_TRACK4_OFF);
   dcMemory.SelectObject(&bmTrack4);
   devc->BitBlt(XOFFSET+(CHANNEL_CX*3), 0, CHANNEL_CX, ROWHEIGHT, &dcMemory, 
      0, 0, SRCCOPY);
   
   dcMemory.SelectObject(pOldBitmap);
  
}




void CChildView::PatEd_DrawPattern(CDC *devc)
{
	char tBuf[16];
   int Row = mFirstRow;
   int yOffset = YOFFSET;
   int Channel = 0;
   
   CRect rClient; GetClientRect(&rClient);
   rClient.top = YOFFSET;
   CBrush c(mcrMain);
   devc->FillRect(&rClient,&c);

   int NumRows = (rClient.bottom - rClient.top) / ROWHEIGHT;

   // initialise scrollbar if necessary
   if (NumRows < PATSIZE)
   {
      ShowScrollBar(SB_VERT,TRUE);
	   SetScrollRange(SB_VERT,0,PATSIZE-NumRows);
      SetScrollPos(SB_VERT, mFirstRow);
   }
   else
   {
      ShowScrollBar(SB_VERT, FALSE);
   }

   // Draw pattern information
   for (int i = 0; i < NumRows; ++i)
   {
      devc->SetTextColor(0);
      if (Row % 4 == 0)
      {
         devc->SetBkColor(mcrMain5);
         CRect rrow(0, yOffset, rClient.right, yOffset+ROWHEIGHT);
         CBrush crow(mcrMain5);
         devc->FillRect(&rrow, &crow);
      }
      else
      {
         devc->SetBkColor(mcrMain);
      }

      // out line number
      sprintf(tBuf,"%02d",Row);
      devc->TextOut(5, yOffset, tBuf);

      for (int ChanNum = 0; ChanNum < MAX_CHANNELS; ChanNum++)
      {
         // out note
         devc->TextOut(XOFFSET+5+(ChanNum*CHANNEL_CX), yOffset, NoteScale[mpCurPattern->GetNote(ChanNum, Row)]);

         // out oct
         if (mpCurPattern->GetOct(ChanNum, Row) == -1)
            sprintf(tBuf,"-");
         else
            sprintf(tBuf,"%d", mpCurPattern->GetOct(ChanNum, Row));
         devc->TextOut(XOFFSET+40+(ChanNum*CHANNEL_CX), yOffset, tBuf);

         // out voice
         if (mpCurPattern->GetVoice(ChanNum, Row) == -1)
            sprintf(tBuf,"- -");
         else
            sprintf(tBuf,"%02d", mpCurPattern->GetVoice(ChanNum, Row));
         devc->TextOut(XOFFSET+65+(ChanNum*CHANNEL_CX), yOffset, tBuf);

         // out effect
         if (mpCurPattern->GetEffect(ChanNum, Row) == -1)
            sprintf(tBuf,"- -");
         else
            sprintf(tBuf,"%02d", mpCurPattern->GetEffect(ChanNum, Row));
         devc->TextOut(XOFFSET+95+(ChanNum*CHANNEL_CX), yOffset, tBuf);

         // out value
         if (mpCurPattern->GetValue(ChanNum, Row) == -1)
            sprintf(tBuf,"- -");
         else
            sprintf(tBuf,"%02d", mpCurPattern->GetValue(ChanNum, Row));
         devc->TextOut(XOFFSET+125+(ChanNum*CHANNEL_CX), yOffset, tBuf);
      }

      Row++;
      yOffset+=ROWHEIGHT;
   }

   for (int ChanNum = 0; ChanNum < MAX_CHANNELS+1; ChanNum++)
   {
      devc->MoveTo(XOFFSET+(ChanNum*CHANNEL_CX), rClient.top);
      devc->LineTo(XOFFSET+(ChanNum*CHANNEL_CX), rClient.bottom);

   }

}




void CChildView::OnVScroll(UINT nSBCode, UINT nPos, CScrollBar* pScrollBar) 
{
   GetParent()->SetFocus(); // first ensure the pattern editor has the focus

   CRect rClient; GetClientRect(&rClient);
   int NumRows = (rClient.bottom - rClient.top - YOFFSET) / ROWHEIGHT;

   switch(nSBCode)
	{
		case SB_LINEDOWN:
         if (mFirstRow + NumRows < PATSIZE)
         {
            mFirstRow++;
            Invalidate();
            if (mCurRow + NumRows < PATSIZE)
               mCurRow++;
            PatEd_DrawCursor();
         }
         break;

		case SB_LINEUP:
         if (mFirstRow > 0)
         {
            mFirstRow--;
            Invalidate();
            if (mCurRow > 0)
               mCurRow--;
            PatEd_DrawCursor();
         }
			break;

      case SB_PAGEDOWN:
         if (mFirstRow + NumRows + 15 < PATSIZE)
         {
            mFirstRow+=16;
            Invalidate();
            if (mCurRow + NumRows + 15 < PATSIZE)
               mCurRow+=16;
            else
               mCurRow = PATSIZE;
            PatEd_DrawCursor();

         }
         break;

      case SB_PAGEUP:
         if (mFirstRow > 15)
         {
            mFirstRow-=16;
            Invalidate();
            if (mCurRow > 15)
               mCurRow-=16;
            else
               mCurRow = 0;
            PatEd_DrawCursor();
         }
			break;

      case SB_THUMBPOSITION:
		case SB_THUMBTRACK:
         {
            int Diff = mCurRow - mFirstRow;
            mFirstRow = nPos;
            Invalidate();
            if (nPos + Diff < PATSIZE)
               mCurRow = nPos + Diff;
            else
               mCurRow = PATSIZE;
            PatEd_DrawCursor();
         }
			break;

      default: 
         break;
	}
}



void CChildView::OnKeyDown(UINT nChar, UINT nRepCnt, UINT nFlags) 
{
   switch (nChar)
   {
   case VK_DOWN:
   case VK_UP:
   case VK_LEFT:
   case VK_RIGHT:
      PatEd_MoveCursor(nChar);
      break;

   case VK_END:
      PatEd_CheckNote(nChar);
      break;
   }
}

void CChildView::OnChar(UINT nChar, UINT nRepCnt, UINT nFlags) 
{
   // calc position
   int Item = mCurX % MAXITEMS;

   switch(Item)
   {
   case 0: // note
      PatEd_CheckNote(nChar); 
      break;

   case 1: // Oct
      PatEd_CheckOct(nChar);
      break;

   case 2: // Voice
      PatEd_CheckVoice(nChar);
      break;

   case 3: // Effect
      PatEd_CheckEffect(nChar);
      break;
   
   case 4: // Value
      PatEd_CheckValue(nChar);
      break;
   }
}


void CChildView::PatEd_CheckNote(UINT nChar) 
{
   const char * Keys;

   if (Global::bQwertyLayout)
      Keys = szQwertyLayout;
   else
      Keys = szAzertyLayout;

   nChar = toupper(nChar);
   
   char * pFound = strchr(Keys, nChar);
   if (pFound == NULL && nChar != VK_END) return; // not found

   int pos;
   if (nChar == VK_END)
      pos = NOTE_END;
   else
      pos = pFound - Keys;

   int ChanNum = mCurX / MAXITEMS;
   mpCurPattern->SetNote(ChanNum, mCurRow, pos);

   int Octave = Global::pMainFrm->ToolBar_GetOctave();
   int Voice = Global::pMainFrm->ToolBar_GetCurrentVoice();

   if (pos == NOTE_BLANK || pos == NOTE_END)
   {
      // need to blank out other values.
      mpCurPattern->SetValue(ChanNum, mCurRow, -1);
      mpCurPattern->SetEffect(ChanNum, mCurRow, -1);
      mpCurPattern->SetOct(ChanNum, mCurRow, -1);
      mpCurPattern->SetVoice(ChanNum, mCurRow, -1);
   }
   else
   {
      // set other fields to default values
      mpCurPattern->SetOct(ChanNum, mCurRow, Octave);
      mpCurPattern->SetVoice(ChanNum, mCurRow, Voice+1);
      mpCurPattern->SetEffect(ChanNum, mCurRow, 0);
      mpCurPattern->SetValue(ChanNum, mCurRow, 0);
      // play note
      KeyUp();
      if (m_LastInstrPlayed != Voice)
      {
         unsigned char Instr[INSTRLEN];
         Global::VoiceArray[Voice].GetInstr(Instr);
         loadInstrument(Instr);
      }
      if (pos == 0) Octave--; // hack to make C first note in octave
      KeyDown(pos, Octave);
   }

   PatEd_RedrawRow(mCurRow,ChanNum);
   PatEd_MoveCursor(VK_DOWN);
}



void CChildView::PatEd_CheckOct(UINT nChar) 
{
   int ChanNum = mCurX / MAXITEMS;

   // first check range
   int Value = nChar - '0';
   if (Value < LOWER_OCT_RANGE || Value > UPPER_OCT_RANGE)
      return;

   mpCurPattern->SetOct(ChanNum, mCurRow, Value);

   PatEd_RedrawRow(mCurRow,ChanNum);
   Global::pMainFrm->ToolBar_SetOctave(Value);
}

void CChildView::PatEd_CheckVoice(UINT nChar) 
{
   int ChanNum = mCurX / MAXITEMS;

   // first check range
   int Value = nChar - '0';
   if (Value < 0 || Value > 9)
      return;

   int CurVal = mpCurPattern->GetVoice(ChanNum, mCurRow);
   switch (CurVal)
   {
   case 0:
      break;
   case -1:
   default:
      // update units
      if (Value != 0)
      CurVal = Value;
      break;

   case 1:
   case 2: 
      // convert units to 10s and add units
      CurVal = (CurVal * 10) + Value;
      break;

   case 3:
      // only values up to 31 allowed - if above assume units
      if (Value == 0 || Value == 1)
         CurVal = (CurVal * 10) + Value;
      else
         CurVal = Value;
      break;

   }

   mpCurPattern->SetVoice(ChanNum, mCurRow, CurVal);

   PatEd_RedrawRow(mCurRow,ChanNum);
   Global::pMainFrm->ToolBar_SetCurrentVoice(CurVal -1);
}

void CChildView::PatEd_CheckEffect(UINT nChar) 
{
   int ChanNum = mCurX / MAXITEMS;

   // first check range
   int Value = nChar - '0';
   if (Value < 0 || Value > 9)
      return;

   int CurVal = mpCurPattern->GetEffect(ChanNum, mCurRow);
   
   switch (CurVal)
   {
   case -1:
   case 0: 
   default:
      // update units
      CurVal = Value;
      break;

   case 1:
      // only values up to 31 allowed - if above assume units
      if (Value < 6)
         CurVal = (CurVal * 10) + Value;
      else
         CurVal = Value;
      break;

   }

   mpCurPattern->SetEffect(ChanNum, mCurRow, CurVal);

   PatEd_RedrawRow(mCurRow,ChanNum);
}

void CChildView::PatEd_CheckValue(UINT nChar) 
{
   int ChanNum = mCurX / MAXITEMS;

   // first check range
   int Value = nChar - '0';
   if (Value < 0 || Value > 9)
      return;

   int CurVal = mpCurPattern->GetValue(ChanNum, mCurRow);
   
   if (CurVal == -1)
      CurVal = Value;
   else
   {
      // shift digit left and provide a number in range 0 - 99
      CurVal *= 10;
      CurVal += Value;
      CurVal %= 100;
   }

   mpCurPattern->SetValue(ChanNum, mCurRow, CurVal);

   PatEd_RedrawRow(mCurRow,ChanNum);
}



void CChildView::PatEd_RedrawRow(int Row, int ChanNum) 
{
   CDC *dc = GetDC();

   PatEd_DrawCursor(); // erase cursor

   CRect rBlank(XOFFSET+1+(ChanNum*CHANNEL_CX),
      YOFFSET+((Row-mFirstRow)*ROWHEIGHT),
      XOFFSET+(ChanNum*CHANNEL_CX)+CHANNEL_CX-1,
      YOFFSET+((Row-mFirstRow)*ROWHEIGHT)+ROWHEIGHT);

   if (mCurRow % 4 == 0)
   {
      CBrush crow(mcrMain5);
      dc->FillRect(&rBlank, &crow);
      dc->SetBkColor(mcrMain5);
   }
   else
   {
      CBrush crow(mcrMain);
      dc->SetBkColor(mcrMain);
      dc->FillRect(&rBlank, &crow);
   }
   
   char tBuf[100];
   int yOffset = YOFFSET + (Row-mFirstRow)*ROWHEIGHT;

   // out note
   dc->TextOut(XOFFSET+5+(ChanNum*CHANNEL_CX), yOffset, 
      NoteScale[mpCurPattern->GetNote(ChanNum,Row)]);

   // out oct
   if (mpCurPattern->GetOct(ChanNum,Row) == -1)
      sprintf(tBuf,"-");
   else
      sprintf(tBuf,"%d", mpCurPattern->GetOct(ChanNum,Row));
   dc->TextOut(XOFFSET+40+(ChanNum*CHANNEL_CX), yOffset, tBuf);

   // out voice
   if (mpCurPattern->GetVoice(ChanNum,Row) == -1)
      sprintf(tBuf,"- -");
   else
      sprintf(tBuf,"%02d", mpCurPattern->GetVoice(ChanNum,Row));
   dc->TextOut(XOFFSET+65+(ChanNum*CHANNEL_CX), yOffset, tBuf);

   // out effect
   if (mpCurPattern->GetEffect(ChanNum,Row) == -1)
      sprintf(tBuf,"- -");
   else
      sprintf(tBuf,"%02d", mpCurPattern->GetEffect(ChanNum,Row));
   dc->TextOut(XOFFSET+95+(ChanNum*CHANNEL_CX), yOffset, tBuf);

   // out value
   if (mpCurPattern->GetValue(ChanNum,Row) == -1)
      sprintf(tBuf,"- -");
   else
      sprintf(tBuf,"%02d", mpCurPattern->GetValue(ChanNum,Row));
   dc->TextOut(XOFFSET+125+(ChanNum*CHANNEL_CX), yOffset, tBuf);

   PatEd_DrawCursor(); // redraw cursor

}





void CChildView::PatEd_MoveCursor(UINT nChar) 
{
   CDC *dc = GetDC();

   CRect rClient; GetClientRect(&rClient);
   int NumRows = (rClient.bottom - rClient.top - YOFFSET) / ROWHEIGHT;

   switch(nChar)
   {
   case VK_DOWN:
      if (mCurRow < PATSIZE-1)
      {
         // need to scroll into view?
         if (mCurRow == mFirstRow + NumRows -1)
         {
            mCurRow++;
            mFirstRow++;
            Invalidate();
         }
         else
         {
   
            PatEd_DrawCursor(); // erase old
            mCurRow++;
            PatEd_DrawCursor(); // draw new
         }  
      }
      break;

   case VK_UP:
      if (mCurRow > 0)
      {
         // need to scroll into view?
         if (mCurRow == mFirstRow)
         {
            mCurRow--;
            mFirstRow--;
            Invalidate();
         }
         else
         {
            PatEd_DrawCursor(); // erase old
            mCurRow--;
            PatEd_DrawCursor(); // draw new
         }  
      }
      break;

   case VK_LEFT:
      if (mCurX > 0)
      {
         PatEd_DrawCursor(); // erase old
         mCurX--;
         PatEd_DrawCursor(); // draw new
      }
      break;

   case VK_RIGHT:
      if (mCurX < (MAX_CHANNELS * MAXITEMS) - 1)
      {
         PatEd_DrawCursor(); // erase old
         mCurX++;
         PatEd_DrawCursor(); // draw new
      }
      break;

   }
}


void CChildView::OnLButtonDown(UINT nFlags, CPoint point) 
{
   GetParent()->SetFocus(); // first ensure the pattern editor has the focus

   // check if channel states need changing
   bool bChannelChanged = false;

   for (int i = 0; i < MAX_CHANNELS; i++)
   {
      CRect rTrack(XOFFSET+(CHANNEL_CX*i), 0, XOFFSET+(CHANNEL_CX*(i+1)), ROWHEIGHT);

      if (PtInRect(rTrack, point)) 
      {
         Global::bChannelActive[i] = !Global::bChannelActive[i];
         bChannelChanged = true;
      }
   }

   if (bChannelChanged)
   {
      CDC * dc = GetDC();
      PatEd_DrawHeader(dc);
      return;
   }

   // check for mouse position in rest of the window
   if (point.x < XOFFSET || point.x > XOFFSET + (CHANNEL_CX*MAX_CHANNELS))
      return; // out of range

   CRect rClient; GetClientRect(&rClient);
   int NumRows = (rClient.bottom - rClient.top - YOFFSET) / ROWHEIGHT;

   int yPos = (point.y - YOFFSET) / ROWHEIGHT;
   if (yPos >= NumRows)
      return; // out of range

   int Channel = (point.x - XOFFSET) / CHANNEL_CX;
   int Item = (point.x - XOFFSET) % CHANNEL_CX;
   int Pos = 0;
   for (i = MAXITEMS-1; i >= 0; i--)
   {
      if (Item > CurOffsets[i * 2])
      {
         Pos = i;
         break;
      }
   }

   PatEd_DrawCursor(); // erase cursor
   mCurRow = mFirstRow + yPos;
   mCurX = (Channel * MAXITEMS) + Pos;
   PatEd_DrawCursor(); // redraw

}


BOOL CChildView::OnEraseBkgnd(CDC* pDC)
{
   return true;
}

void CChildView::PatEd_OnPatternPlay() 
{

}





// --------------------------------------------------------
// Track Octave modifiers
//
//
// --------------------------------------------------------

// Decrease the volume value for all of the notes in the current track
// and pattern. 
void CChildView::OnEditDecreasetrackoctave() 
{
   int i;

   int ChanNum = mCurX / MAXITEMS;
   int PatNum = ((CMainFrame *)GetParent())->ToolBar_GetPattern();

   // first do a sanity check to ensure that we won't lose
   // any notes by doing this - all notes are oct 2+
   char lowestOct = 7;
   for (i = 0; i < PATSIZE; i++)
   {
      char Oct = Global::Patterns[PatNum].GetOct(ChanNum, i);
      if (Oct != -1 && Oct < lowestOct) lowestOct = Oct;
   }

   // WARN
   if (lowestOct == LOWER_OCT_RANGE)
   {
      if (AfxMessageBox("Some notes will be lost, continue?", MB_YESNO) == IDNO)
         return;
   }

   for (i = 0; i < PATSIZE; i++)
   {
      char Oct = Global::Patterns[PatNum].GetOct(ChanNum, i);

      if (Oct > LOWER_OCT_RANGE) Oct --;
      Global::Patterns[PatNum].SetOct(ChanNum, i, Oct);
   }

   // force pattern redraw
   Invalidate();
	
}

void CChildView::OnEditIncreasetrackoctave() 
{
   int i;
   int ChanNum = mCurX / MAXITEMS;
   int PatNum = ((CMainFrame *)GetParent())->ToolBar_GetPattern();

   // first do a sanity check to ensure that we won't lose
   // any notes by doing this - all notes are oct 2+
   char highestOct = 0;
   for (i = 0; i < PATSIZE; i++)
   {
      char Oct = Global::Patterns[PatNum].GetOct(ChanNum, i);
      if (Oct != -1 && Oct > highestOct) highestOct = Oct;
   }

   // WARN
   if (highestOct == UPPER_OCT_RANGE)
   {
      if (AfxMessageBox("Some notes will be lost, continue?", MB_YESNO) == IDNO)
         return;
   }

   // clear current track
   for (i = 0; i < PATSIZE; i++)
   {
      char Oct = Global::Patterns[PatNum].GetOct(ChanNum, i);

      if (Oct > 0 && Oct < UPPER_OCT_RANGE) Oct ++;
      Global::Patterns[PatNum].SetOct(ChanNum, i, Oct);
   }

   // force pattern redraw
   Invalidate();
}


// --------------------------------------------------------
// Song Building Code
//
//
// --------------------------------------------------------


// scans the entire song for the largest instrument number
// as this is the value we have to send to the driver to 
// save a few bytes rather than sending 32 instruments all of
// the time
int CChildView::GetHighestInstrNo()
{
   int HighestInstrNo = 1;

   for (int Pats = 0; Pats < Global::NumPatterns; Pats++)
   {
      CPattern * pPat = &Global::Patterns[Pats];
      int InstrNo = pPat->GetHighestInstrNo();

      if (InstrNo > HighestInstrNo)
         HighestInstrNo = InstrNo;
   }
    
   return HighestInstrNo;
}




void CChildView::OnPhButPlay() { OnPlayCurrentpattern(); }
void CChildView::OnPlayCurrentpattern() 
{
   int Len = sizeof(Z80ModDriver);
   memcpy(SongBuffer, Z80ModDriver, Len);

   // add a dummy V1 SampleTable
   Len+=BYTESIZE_SAMPLETABLE;

   memcpy(&SongBuffer[Len], "MVS0", 4); // header
   Len+=4;
   // offset to next song, but there is only 1
   SongBuffer[Len++] = 0;
   SongBuffer[Len++] = 0;

   // number of instruments
   int NumInstrs = GetHighestInstrNo();
   SongBuffer[Len++] = (char) NumInstrs;
   
   for (int i = 1; i <= NumInstrs; i ++)
   {
      unsigned char Instr[INSTRLEN];
      Global::VoiceArray[i].GetInstr(Instr);
      memcpy(&SongBuffer[Len], Instr, INSTRLEN);
      Len+=INSTRLEN;
      memset(&SongBuffer[Len], 0, 7); // instr padding
      Len+=7;
   }

   // song length
   SongBuffer[Len++] = 1;
   // pattern order
   memset(&SongBuffer[Len], 0, 128);
   Len+= 128;

   memcpy(&SongBuffer[Len], mpCurPattern->BuildPattern(), BUILTPATSIZE);
   Len+= BUILTPATSIZE;

   sendData(SongBuffer, Len);
   Sleep(100); // wait for z80 to start
   sendCommand(0x81); // send command to play first FM tune.
}


int CChildView::BuildSong() 
{
   int i;
   int Len = sizeof(Z80ModDriver);
   memcpy(SongBuffer, Z80ModDriver, Len);
   
   // add a dummy V1 SampleTable
   Len+=BYTESIZE_SAMPLETABLE;

   memcpy(&SongBuffer[Len], "MVS0", 4); // header
   Len+=4;
   // offset to next song, but there is only 1
   SongBuffer[Len++] = 0;
   SongBuffer[Len++] = 0;

   // number of instruments
   int NumInstrs = GetHighestInstrNo();
   SongBuffer[Len++] = (char) NumInstrs;
   
   for (i = 1; i <= NumInstrs; i ++)
   {
      unsigned char Instr[INSTRLEN];
      Global::VoiceArray[i].GetInstr(Instr);
      memcpy(&SongBuffer[Len], Instr, INSTRLEN);
      Len+=INSTRLEN;
      memset(&SongBuffer[Len], 0, 7); // instr padding
      Len+=7;
   }

   // song length
   int SongLen = Global::pMainFrm->ToolBar_GetSongLen();
   SongBuffer[Len++] = (char)SongLen;

   for (i = 0; i < SongLen; i ++)
   {
      SongBuffer[Len++] = (char)Global::pMainFrm->ToolBar_GetSongPattern(i);
   }
   
   // pad the rest of the song out
   while (i < 128)
   {
      SongBuffer[Len++] = 0;
      i++;
   }

   // now store all of the patterns
   for (i = 0; i < Global::NumPatterns; i ++)
   {  
      CPattern * pPat = &Global::Patterns[i];
      memcpy(&SongBuffer[Len], pPat->BuildPattern(), BUILTPATSIZE);
      Len+= BUILTPATSIZE;
   }

   return Len;
}

void CChildView::OnPhButPlayAll() { OnPlaySong(); }
void CChildView::OnPlaySong() 
{
   int Len = BuildSong();
   sendData(SongBuffer, Len);
   Sleep(100); // wait for z80 to start
   sendCommand(0x81); // send command to play first FM tune.
}

void CChildView::OnPhButStop() { OnPlayStop(); }
void CChildView::OnPlayStop() 
{
   sendCommand(0xff); // send command to stop all FM
}

void CChildView::OnFileExporttestrom() 
{
   int Len = BuildSong();

   // hack byte 4 so that the song plays instantly without the
   // 68000 having to send the play command to the z80
   SongBuffer[4] = 1;

   CFile fWorkspace;
   char NullChar = 0;
   char CurDir[MAX_PATH+1];
   char Filename[MAX_PATH+1] = "202-m1";

   CFileDialog dlgFileSave(false, "BIN", NULL, OFN_HIDEREADONLY | OFN_OVERWRITEPROMPT,
      szTestROMFilter, NULL);
   GetCurrentDirectory(MAX_PATH, CurDir);
   dlgFileSave.m_ofn.lpstrInitialDir = CurDir;
   dlgFileSave.m_ofn.lpstrFile = Filename;
   dlgFileSave.m_ofn.nMaxFile = MAX_PATH;
   if (dlgFileSave.DoModal() == IDCANCEL)
      return;
   
   if (!fWorkspace.Open(dlgFileSave.GetPathName(), CFile::modeWrite | CFile::modeCreate))
   {
      CString ErrMsg;
      ErrMsg = "Failed to Create " + dlgFileSave.GetFileName();
      AfxMessageBox(ErrMsg);
      return;
   }

   fWorkspace.Write(SongBuffer, Len);
   fWorkspace.Close();
}

#include "mdTestRom.h"

void CChildView::OnFileExportMdTestRom()
{
	int Len = BuildSong();
	int Index=sizeof(Z80ModDriver) + BYTESIZE_SAMPLETABLE;
	int RomLen=0;
	Len-=Index;
	unsigned char *Data=(SongBuffer+Index);
	
	CFile fWorkspace;
	char NullChar = 0;
	char CurDir[MAX_PATH+1];
	char Filename[MAX_PATH+1] = "testrom";

	CFileDialog dlgFileSave(false, "BIN", NULL, OFN_HIDEREADONLY | OFN_OVERWRITEPROMPT,
	  szTestROMFilter, NULL);
	GetCurrentDirectory(MAX_PATH, CurDir);
	dlgFileSave.m_ofn.lpstrInitialDir = CurDir;
	dlgFileSave.m_ofn.lpstrFile = Filename;
	dlgFileSave.m_ofn.nMaxFile = MAX_PATH;
	if (dlgFileSave.DoModal() == IDCANCEL)
	  return;

	if (!fWorkspace.Open(dlgFileSave.GetPathName(), CFile::modeWrite | CFile::modeCreate))
	{
	  CString ErrMsg;
	  ErrMsg = "Failed to Create " + dlgFileSave.GetFileName();
	  AfxMessageBox(ErrMsg);
	  return;
	}

	RomLen=sizeof(MdTestRom);
	fWorkspace.Write(MdTestRom,sizeof(MdTestRom));
	RomLen+=Len;
	fWorkspace.Write(Data,Len);

	// pad
	int pad_size=(1024*128) - RomLen;
	unsigned char *pad=(unsigned char*)malloc(pad_size);
	
	memset(pad,0,pad_size);
	fWorkspace.Write(pad,pad_size);
	free(pad);


	fWorkspace.Close();
}


void CChildView::OnFileExportMVS0()
{
	int Len = BuildSong();
	int Index=sizeof(Z80ModDriver) + BYTESIZE_SAMPLETABLE;
	Len-=Index;
	unsigned char *Data=(SongBuffer+Index);
	
	// Save file
	CFile fWorkspace;
	char NullChar = 0;
	char CurDir[MAX_PATH+1];
	char Filename[MAX_PATH+1] = "";

	CFileDialog dlgFileSave(false, "mv0", NULL, OFN_HIDEREADONLY | OFN_OVERWRITEPROMPT,
	  szMVS0ROMFilter, NULL);
	GetCurrentDirectory(MAX_PATH, CurDir);
	dlgFileSave.m_ofn.lpstrInitialDir = CurDir;
	dlgFileSave.m_ofn.lpstrFile = Filename;
	dlgFileSave.m_ofn.nMaxFile = MAX_PATH;
	if (dlgFileSave.DoModal() == IDCANCEL)
	  return;

	if (!fWorkspace.Open(dlgFileSave.GetPathName(), CFile::modeWrite | CFile::modeCreate))
	{
	  CString ErrMsg;
	  ErrMsg = "Failed to Create " + dlgFileSave.GetFileName();
	  AfxMessageBox(ErrMsg);
	  return;
	}

	fWorkspace.Write(Data, Len);
	fWorkspace.Close();

}


void CChildView::OnFileSave() 
{
   CFile fWorkspace;
   char NullChar = 0;
   char CurDir[MAX_PATH+1];

   CFileDialog dlgFileSave(false, "MUS", NULL, OFN_HIDEREADONLY | OFN_OVERWRITEPROMPT,
      szMusicFilter, NULL);
   GetCurrentDirectory(MAX_PATH, CurDir);
   dlgFileSave.m_ofn.lpstrInitialDir = CurDir;
   dlgFileSave.m_ofn.nMaxFile = MAX_PATH;
   if (dlgFileSave.DoModal() == IDCANCEL)
      return;
   
   if (!fWorkspace.Open(dlgFileSave.GetPathName(), CFile::modeWrite | CFile::modeCreate))
   {
      CString ErrMsg;
      ErrMsg = "Failed to Create " + dlgFileSave.GetFileName();
      AfxMessageBox(ErrMsg);
      return;
   }

   // write header 'MVSM' + ver num of save format
   CString Header(szMusicHeader);
   fWorkspace.Write(Header, sizeof(szMusicHeader)-1);

   // save all instruments
   int i;
   for (i = 1; i < MAX_VOICES; ++i)
   {
      unsigned char Instr[INSTRLEN];
      CString VoiceName = Global::VoiceArray[i].GetName();
      VoiceName = VoiceName.Right(VoiceName.GetLength() - 3);
      fWorkspace.Write(VoiceName, VoiceName.GetLength());
      fWorkspace.Write(&NullChar, 1); // null terminate string
      Global::VoiceArray[i].GetInstr(Instr);
      fWorkspace.Write(Instr, INSTRLEN);
   }

   // save song song len followed by pattern order
   char SongLen = (char)Global::pMainFrm->ToolBar_GetSongLen();
   fWorkspace.Write(&SongLen, 1);

   char SongByte;
   for (i = 0; i < SongLen; i ++)
   {
      SongByte = (char)Global::pMainFrm->ToolBar_GetSongPattern(i);
      fWorkspace.Write(&SongByte, 1);
   }
   
   // save song #patterns followed by patterns
   char NumPats = Global::NumPatterns;
   fWorkspace.Write(&NumPats, 1);
   for (i = 0; i < Global::NumPatterns; i ++)
   {  
      CPattern * pPat = &Global::Patterns[i];

      CString PatName = pPat->GetPatternName();
      fWorkspace.Write(PatName, PatName.GetLength());
      fWorkspace.Write(&NullChar, 1); // write null terminator
      fWorkspace.Write(pPat->BuildPattern(), BUILTPATSIZE);
   }

   fWorkspace.Close();
}

void CChildView::OnFileNewproject() 
{
   int i;

   Global::NumPatterns = 0;

   // Give different names to voices on start up
   for (i = 1; i < MAX_VOICES; i++)
   {
      CString VName;
      VName.Format("%02d Unnamed", i);
      Global::VoiceArray[i].SetName(VName);
      Global::VoiceArray[i].Clear();
   }

   for (i = 0; i < MAX_CHANNELS; i++)
      Global::bChannelActive[i] = true;

   for (i = 0; i < MAX_PATTERNS; i++)
      Global::Patterns[i].Clear();

   Global::pMainFrm->InitialiseToolBars();
}

void CChildView::OnFileOpen() 
{
   // first clear the current project
   OnFileNewproject();

   // now load the new one
   CFile fWorkspace;
   char NullChar = 0;
   char CurDir[MAX_PATH+1];

   CFileDialog dlgFileOpen(true, "MUS", NULL, OFN_HIDEREADONLY | OFN_PATHMUSTEXIST | OFN_FILEMUSTEXIST,
      szMusicFilter, NULL);
   GetCurrentDirectory(MAX_PATH, CurDir);
   dlgFileOpen.m_ofn.lpstrInitialDir = CurDir;
   if (dlgFileOpen.DoModal() == IDCANCEL)
      return;
	
   if (!fWorkspace.Open(dlgFileOpen.GetPathName(), CFile::modeRead))
   {
      CString ErrMsg;
      ErrMsg = "Failed to Open " + dlgFileOpen.GetFileName();
      AfxMessageBox(ErrMsg);
      return;
   }
      
   // validate header
   char Header[sizeof(szMusicHeader)];
   fWorkspace.Read(Header, sizeof(szMusicHeader)-1);
   if (memcmp(Header, szMusicHeader, sizeof(szMusicHeader)-1) != 0)
   {
      AfxMessageBox("Invalid MVS Tracker Music file");
      return;
   }

   // read in all of the instruments
   int i;
   for (i = 1; i < MAX_VOICES; ++i)
   {
      // get voice name and data
      CString RawInstrName = Global::ReadNullTerminatedString(&fWorkspace);
      CString InstrName;
      InstrName.Format("%02d %s", i, RawInstrName);
      unsigned char Instr[INSTRLEN];

      fWorkspace.Read(Instr, INSTRLEN);
      // store
      Global::VoiceArray[i].SetName(InstrName);
      Global::VoiceArray[i].SetInstr(Instr);
   }

   // update main window's combo box 
   Global::pMainFrm->ToolBar_FillCombo();
   
   // load song: song len followed by pattern order
   char SongLen;
   fWorkspace.Read(&SongLen, 1);

   char SongByte;
   Global::pMainFrm->ToolBar_ClearSongPatterns();
   for (i = 0; i < SongLen; i ++)
   {
      fWorkspace.Read(&SongByte, 1);
      Global::pMainFrm->ToolBar_AddSongPattern(SongByte);
   }
   
   // read in number of patterns
   char NumPats;
   fWorkspace.Read(&NumPats, 1);
   Global::NumPatterns = NumPats;
   
   // now store all of the patterns
   for (i = 0; i < Global::NumPatterns; i ++)
   {  
      CPattern * pPat = &Global::Patterns[i];

      // get voice name and data
      CString PatName = Global::ReadNullTerminatedString(&fWorkspace);
      pPat->SetPatternName(PatName);

      unsigned char NewPat[BUILTPATSIZE];
      fWorkspace.Read(NewPat, BUILTPATSIZE);
      pPat->SetPattern(NewPat);
   }

   fWorkspace.Close();

   Global::pMainFrm->ToolBar_RefreshPatName();
}


void CChildView::OnFileImportmod() 
{
   // first clear the current project
   OnFileNewproject();

   // now load the new one
   CFile fWorkspace;
   char NullChar = 0;
   char CurDir[MAX_PATH+1];
   int i;

   CFileDialog dlgFileOpen(true, "MOD", NULL, OFN_HIDEREADONLY | OFN_PATHMUSTEXIST | OFN_FILEMUSTEXIST,
      szMODFilter, NULL);
   GetCurrentDirectory(MAX_PATH, CurDir);
   dlgFileOpen.m_ofn.lpstrInitialDir = CurDir;
   if (dlgFileOpen.DoModal() == IDCANCEL)
      return;
	
   if (!fWorkspace.Open(dlgFileOpen.GetPathName(), CFile::modeRead))
   {
      CString ErrMsg;
      ErrMsg = "Failed to Open " + dlgFileOpen.GetFileName();
      AfxMessageBox(ErrMsg);
      return;
   }
      
   // validate header
   char Header[sizeof(szMODHeader)];
   fWorkspace.Seek(MOD_HEADER, 0);
   fWorkspace.Read(Header, sizeof(szMODHeader)-1);
   if (memcmp(Header, szMODHeader, sizeof(szMODHeader)-1) != 0)
   {
      AfxMessageBox("Invalid or Incompatible MOD file");
      return;
   }

   // load song: song len followed by pattern order
   fWorkspace.Seek(MOD_SONG_LENGTH, 0);
   char SongLen;
   fWorkspace.Read(&SongLen, 1);

   char SongByte;
   fWorkspace.Read(&SongByte, 1); // skip CIAA

   // read in song table
   Global::pMainFrm->ToolBar_ClearSongPatterns();
   for (i = 0; i < SongLen; i ++)
   {
      fWorkspace.Read(&SongByte, 1);
      Global::pMainFrm->ToolBar_AddSongPattern(SongByte);
   }
   
   // calculate number of patterns
   int TotalInstrLen = 0;
   BYTE readByte;

   for (i = 0; i < 31; i ++)
   {
      int instrLen;
      fWorkspace.Seek(i*30+42, 0);
      fWorkspace.Read(&readByte, 1);
      instrLen = readByte << 8;
      fWorkspace.Read(&readByte, 1);
      instrLen += readByte;
      instrLen *= 2;
      TotalInstrLen += instrLen;
   }

   DWORD FileLen = fWorkspace.GetLength();
   Global::NumPatterns = (FileLen - MOD_PATTERN_START - TotalInstrLen) / 1024;

   int Patterns, Row, Channel;

   for (Patterns = 0; Patterns < Global::NumPatterns; Patterns++)
   {
      CPattern * pPat = &Global::Patterns[Patterns];
      CString PatName;
      PatName.Format("Pat %02d", Patterns);
      pPat->SetPatternName(PatName);

      fWorkspace.Seek((Patterns * 1024) + MOD_PATTERN_START, 0);
      int LastChannelInstr[4];
      for (i = 0; i < 4; i ++)
         LastChannelInstr[i] = -1;
      

      for (Row = 0; Row < 64; Row++)
      {
         for (Channel = 0; Channel < 4; Channel++)
         {
            int Pitch, Note, Oct;
            int Effect, EffectValue;
            int Instrument;

            // getpitch + hi nibble of instrument
            fWorkspace.Read(&readByte, 1);
            Pitch = readByte;
            Instrument = Pitch & 0xf0;
            Pitch &= 15;
            Pitch <<= 8;
            fWorkspace.Read(&readByte, 1);
            Pitch |= readByte;

            ConvertPitchValue(Pitch, Patterns, Row, Channel, &Note, &Oct);

            fWorkspace.Read(&readByte, 1);
            Effect = readByte; // get instrument and effect number
            Instrument |=  (Effect>>4);
            Effect &= 15; // scrap instrument
            fWorkspace.Read(&readByte, 1);
            EffectValue = readByte;

            if (Note == NOTE_BLANK)
            {
               Oct = -1;
               Instrument = -1;
            }
            else
            {
               if (LastChannelInstr[Channel] == -1)
                  LastChannelInstr[Channel] = Instrument;

               if (Instrument == 0)
                  Instrument = LastChannelInstr[Channel];
            }

            if (Effect == 0)
            {
               Effect = -1;
               EffectValue = -1;
            }

            pPat->SetNote(Channel, Row, Note);
            pPat->SetOct(Channel, Row, Oct);
            pPat->SetVoice(Channel, Row, Instrument);
            pPat->SetEffect(Channel, Row, Effect);
            pPat->SetValue(Channel, Row, EffectValue);
         }
      }
   }

   Global::pMainFrm->ToolBar_RefreshPatName();
   fWorkspace.Close();
}


void CChildView::ConvertPitchValue(int Pitch, int Pattern, int Row, int Channel, 
                                   int * pNote, int * pOct)
{
   int i;

   *pNote = NOTE_BLANK;
   *pOct = -1;
   
   if (Pitch == 0)
      return;      

   for (i = 0; i < sizeof(MOD_Periods); i ++)
   {
      if (MOD_Periods[i] == Pitch)
      {
         *pNote = MOD_Notes[i];   // found
         *pOct = MOD_Octaves[i];
         return;
      }
   }
}


