// InstrEdit.cpp : implementation file
//

#include "stdafx.h"
#include "MVSTracker.h"
#include "InstrEdit.h"
#include "InstrRename.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

/////////////////////////////////////////////////////////////////////////////
// CInstrEdit dialog



CInstrEdit::CInstrEdit(CWnd* pParent /*=NULL*/)
	: CDialog(CInstrEdit::IDD, pParent)
{
	//{{AFX_DATA_INIT(CInstrEdit)
		// NOTE: the ClassWizard will add member initialization here
	//}}AFX_DATA_INIT

   m_pParent = pParent;

   Create(IDD_DLG_INSTR_EDIT);

   InitialiseCombo();

   SetDlgItemInt(IDC_IE_OCTAVE, 4); // set initial octave to 4

   memset(m_CurrentInstr, 0, INSTRLEN);
}


void CInstrEdit::DoDataExchange(CDataExchange* pDX)
{
	CDialog::DoDataExchange(pDX);
	//{{AFX_DATA_MAP(CInstrEdit)
		// NOTE: the ClassWizard will add DDX and DDV calls here
	//}}AFX_DATA_MAP
}


BEGIN_MESSAGE_MAP(CInstrEdit, CDialog)
	//{{AFX_MSG_MAP(CInstrEdit)
	ON_BN_CLICKED(IDC_IE_SCALE, OnIeScale)
	ON_BN_CLICKED(IDC_IE_SILENCE, OnIeSilence)
	ON_BN_CLICKED(IDC_IE_RENAME, OnIeRename)
	ON_BN_CLICKED(IDC_IE_LOAD, OnIeLoad)
	ON_BN_CLICKED(IDC_IE_SAVE, OnIeSave)
	ON_BN_CLICKED(IDC_IE_CLOSE, OnIeClose)
	ON_BN_CLICKED(IDC_IE_STORE, OnIeStore)
	ON_CBN_SELCHANGE(IDC_IE_PRESET_LIST, OnSelchangeIePresetList)
	ON_BN_CLICKED(IDC_IE_CLEAR, OnIeClear)
	ON_WM_SHOWWINDOW()
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// CInstrEdit message handlers

const int cKeyPositions[] = {0,2,4,5,7,9,11,11};
void CInstrEdit::OnIeScale() 
{
   BuildandCheckInstr(); // see if a new instr needs loading	
   CPoint cptMouse; 
   GetCursorPos(&cptMouse); 
   CStatic * bmpScale = (CStatic*) GetDlgItem(IDC_IE_SCALE);
   bmpScale->ScreenToClient(&cptMouse);	

   // black keys
   CRect csharp(5,0,11,24);
   CRect dsharp(13,0,20,24);
   CRect fsharp(29,0,35,24);
   CRect gsharp(37,0,43,24);
   CRect asharp(45,0,51,24);

   int Key = 0;

   // check for mouse in black keys
   if (PtInRect(csharp, cptMouse)) Key = 1;
   else if (PtInRect(dsharp, cptMouse)) Key = 3;
   else if (PtInRect(fsharp, cptMouse)) Key = 6;
   else if (PtInRect(gsharp, cptMouse)) Key = 8;
   else if (PtInRect(asharp, cptMouse)) Key = 10;
   else
   {
      // if no then it must be a white key
      Key = cKeyPositions[cptMouse.x / 8];
   }

   int Octave = GetDlgItemInt(IDC_IE_OCTAVE);
   if (Key == 0 && Octave > 0) Octave --; // hack to make C the first key
   Octave &= 7; // ensure in range

   KeyDown(Key, Octave);
}


void CInstrEdit::OnIeSilence() 
{
   KeyUp();	
}



void CInstrEdit::OnIeClose() 
{
   // check if we need to store first
   unsigned char Instr[INSTRLEN];
   Global::VoiceArray[m_CurrentSelection+1].GetInstr(Instr);

   // build one from the current params
   BuildInstrArray(m_CurrentInstr);
      
   // compare
   if (memcmp(Instr, m_CurrentInstr, INSTRLEN) != 0)
   {
      CString Msg;
      Msg.Format("Store changes to Instrument %d first?", m_CurrentSelection+1);
      if (AfxMessageBox(Msg, MB_YESNO) == IDYES)
         OnIeStore();
   }

   ShowWindow(SW_HIDE);
   
   CMenu* pMenu = m_pParent->GetMenu();
   pMenu->CheckMenuItem(ID_VIEW_INSTR_EDIT, MF_UNCHECKED | MF_BYCOMMAND);

	CDialog::OnOK();
}



/////////////////////////////////////////////////////////////////////////////
// Helper functions

// fills the combo with the preset instruments
void CInstrEdit::InitialiseCombo()
{
   CComboBox * cboPresets = (CComboBox *) GetDlgItem(IDC_IE_PRESET_LIST);

   cboPresets->ResetContent();

   for (int i = 1; i < MAX_VOICES; i ++)
      cboPresets->AddString(Global::VoiceArray[i].GetName());

   cboPresets->SetCurSel(0);
   m_CurrentSelection = 0;
}



// fills edit boxes with presets
void CInstrEdit::PopulateInstr(unsigned char Instr[INSTRLEN])
{
   int loop = 0;
   int val;

   // detune/multiple
   val = Instr[loop++];
   SetDlgItemInt(IDC_IE_DETUNE1, (val>>4));
   SetDlgItemInt(IDC_IE_MULTIPLE1, (val&15));
   val = Instr[loop++];
   SetDlgItemInt(IDC_IE_DETUNE2, (val>>4));
   SetDlgItemInt(IDC_IE_MULTIPLE2, (val&15));
   val = Instr[loop++];
   SetDlgItemInt(IDC_IE_DETUNE3, (val>>4));
   SetDlgItemInt(IDC_IE_MULTIPLE3, (val&15));
   val = Instr[loop++];
   SetDlgItemInt(IDC_IE_DETUNE4, (val>>4));
   SetDlgItemInt(IDC_IE_MULTIPLE4, (val&15));

   // total level
   val = Instr[loop++];
   SetDlgItemInt(IDC_IE_TOTLEVEL1, val);
   val = Instr[loop++];
   SetDlgItemInt(IDC_IE_TOTLEVEL2, val);
   val = Instr[loop++];
   SetDlgItemInt(IDC_IE_TOTLEVEL3, val);
   val = Instr[loop++];
   SetDlgItemInt(IDC_IE_TOTLEVEL4, val);

   // key scale/attack
   val = Instr[loop++];
   SetDlgItemInt(IDC_IE_KEYSCALE1, (val>>6));
   SetDlgItemInt(IDC_IE_ATTACK1, (val&31));
   val = Instr[loop++];
   SetDlgItemInt(IDC_IE_KEYSCALE2, (val>>6));
   SetDlgItemInt(IDC_IE_ATTACK2, (val&31));
   val = Instr[loop++];
   SetDlgItemInt(IDC_IE_KEYSCALE3, (val>>6));
   SetDlgItemInt(IDC_IE_ATTACK3, (val&31));
   val = Instr[loop++];
   SetDlgItemInt(IDC_IE_KEYSCALE4, (val>>6));
   SetDlgItemInt(IDC_IE_ATTACK4, (val&31));

   // Decay
   val = Instr[loop++];
   SetDlgItemInt(IDC_IE_DECAY1, val);
   val = Instr[loop++];
   SetDlgItemInt(IDC_IE_DECAY2, val);
   val = Instr[loop++];
   SetDlgItemInt(IDC_IE_DECAY3, val);
   val = Instr[loop++];
   SetDlgItemInt(IDC_IE_DECAY4, val);

   // Sustain
   val = Instr[loop++];
   SetDlgItemInt(IDC_IE_SUSTAIN1, val);
   val = Instr[loop++];
   SetDlgItemInt(IDC_IE_SUSTAIN2, val);
   val = Instr[loop++];
   SetDlgItemInt(IDC_IE_SUSTAIN3, val);
   val = Instr[loop++];
   SetDlgItemInt(IDC_IE_SUSTAIN4, val);

   // SusLevel/Release
   val = Instr[loop++];
   SetDlgItemInt(IDC_IE_SUSLEVEL1, (val>>4));
   SetDlgItemInt(IDC_IE_RELEASE1, (val&15));
   val = Instr[loop++];
   SetDlgItemInt(IDC_IE_SUSLEVEL2, (val>>4));
   SetDlgItemInt(IDC_IE_RELEASE2, (val&15));
   val = Instr[loop++];
   SetDlgItemInt(IDC_IE_SUSLEVEL3, (val>>4));
   SetDlgItemInt(IDC_IE_RELEASE3, (val&15));
   val = Instr[loop++];
   SetDlgItemInt(IDC_IE_SUSLEVEL4, (val>>4));
   SetDlgItemInt(IDC_IE_RELEASE4, (val&15));

   // feedback/connection
   val = Instr[loop];
   SetDlgItemInt(IDC_IE_FEEDBACK, (val>>3));
   SetDlgItemInt(IDC_IE_CONNECTION, (val&7));
}


void CInstrEdit::BuildInstrArray(unsigned char * DisplayInstr)
{
   int loop = 0;
   int val, val2;

   // detune/multiple
   val = GetDlgItemInt(IDC_IE_DETUNE1);
   val2 = GetDlgItemInt(IDC_IE_MULTIPLE1);
   DisplayInstr[loop++] = ((val & 7) << 4) | (val2 & 15);
   val = GetDlgItemInt(IDC_IE_DETUNE2);
   val2 = GetDlgItemInt(IDC_IE_MULTIPLE2);
   DisplayInstr[loop++] = ((val & 7) << 4) | (val2 & 15);
   val = GetDlgItemInt(IDC_IE_DETUNE3);
   val2 = GetDlgItemInt(IDC_IE_MULTIPLE3);
   DisplayInstr[loop++] = ((val & 7) << 4) | (val2 & 15);
   val = GetDlgItemInt(IDC_IE_DETUNE4);
   val2 = GetDlgItemInt(IDC_IE_MULTIPLE4);
   DisplayInstr[loop++] = ((val & 7) << 4) | (val2 & 15);

   // Total Level
   val = GetDlgItemInt(IDC_IE_TOTLEVEL1);
   DisplayInstr[loop++] = val & 127;
   val = GetDlgItemInt(IDC_IE_TOTLEVEL2);
   DisplayInstr[loop++] = val & 127;
   val = GetDlgItemInt(IDC_IE_TOTLEVEL3);
   DisplayInstr[loop++] = val & 127;
   val = GetDlgItemInt(IDC_IE_TOTLEVEL4);
   DisplayInstr[loop++] = val & 127;

   // keyscale/attack
   val = GetDlgItemInt(IDC_IE_KEYSCALE1);
   val2 = GetDlgItemInt(IDC_IE_ATTACK1);
   DisplayInstr[loop++] = ((val & 3) << 6) | (val2 & 31);
   val = GetDlgItemInt(IDC_IE_KEYSCALE2);
   val2 = GetDlgItemInt(IDC_IE_ATTACK2);
   DisplayInstr[loop++] = ((val & 3) << 6) | (val2 & 31);
   val = GetDlgItemInt(IDC_IE_KEYSCALE3);
   val2 = GetDlgItemInt(IDC_IE_ATTACK3);
   DisplayInstr[loop++] = ((val & 3) << 6) | (val2 & 31);
   val = GetDlgItemInt(IDC_IE_KEYSCALE4);
   val2 = GetDlgItemInt(IDC_IE_ATTACK4);
   DisplayInstr[loop++] = ((val & 3) << 6) | (val2 & 31);

   // Decay
   val = GetDlgItemInt(IDC_IE_DECAY1);
   DisplayInstr[loop++] = val & 31;
   val = GetDlgItemInt(IDC_IE_DECAY2);
   DisplayInstr[loop++] = val & 31;
   val = GetDlgItemInt(IDC_IE_DECAY3);
   DisplayInstr[loop++] = val & 31;
   val = GetDlgItemInt(IDC_IE_DECAY4);
   DisplayInstr[loop++] = val & 31;

   // Sustain
   val = GetDlgItemInt(IDC_IE_SUSTAIN1);
   DisplayInstr[loop++] = val & 31;
   val = GetDlgItemInt(IDC_IE_SUSTAIN2);
   DisplayInstr[loop++] = val & 31;
   val = GetDlgItemInt(IDC_IE_SUSTAIN3);
   DisplayInstr[loop++] = val & 31;
   val = GetDlgItemInt(IDC_IE_SUSTAIN4);
   DisplayInstr[loop++] = val & 31;

   // Sustain Level / Release
   val = GetDlgItemInt(IDC_IE_SUSLEVEL1);
   val2 = GetDlgItemInt(IDC_IE_RELEASE1);
   DisplayInstr[loop++] = ((val & 15) << 4) | (val2 & 15);
   val = GetDlgItemInt(IDC_IE_SUSLEVEL2);
   val2 = GetDlgItemInt(IDC_IE_RELEASE2);
   DisplayInstr[loop++] = ((val & 15) << 4) | (val2 & 15);
   val = GetDlgItemInt(IDC_IE_SUSLEVEL3);
   val2 = GetDlgItemInt(IDC_IE_RELEASE3);
   DisplayInstr[loop++] = ((val & 15) << 4) | (val2 & 15);
   val = GetDlgItemInt(IDC_IE_SUSLEVEL4);
   val2 = GetDlgItemInt(IDC_IE_RELEASE4);
   DisplayInstr[loop++] = ((val & 15) << 4) | (val2 & 15);

   // Feedback / Connection
   val = GetDlgItemInt(IDC_IE_FEEDBACK);
   val2 = GetDlgItemInt(IDC_IE_CONNECTION);
   DisplayInstr[loop] = ((val & 7) << 3) | (val2 & 7);
}


void CInstrEdit::BuildandCheckInstr()
{
   unsigned char DisplayInstr[INSTRLEN];

   BuildInstrArray(DisplayInstr);

   // compare to see if there is any change
   bool bDiff = false;
   int i;
   for (i = 0; i < INSTRLEN; i ++)
   {
      if (DisplayInstr[i] != m_CurrentInstr[i])
      {
         bDiff = true;
         break;
      }
   }

   // if different then load the instrument into YM and update current
   if (bDiff)
   {
      for (i = 0; i < INSTRLEN; i ++)
         m_CurrentInstr[i] = DisplayInstr[i];

      loadInstrument(m_CurrentInstr);
   }
}


// public function called when the dialog is shown
// in order to display the correct instrument
void CInstrEdit::UpdateSelection(int Selection)
{
   CComboBox * cboPresets = (CComboBox *) GetDlgItem(IDC_IE_PRESET_LIST);
   cboPresets->SetCurSel(Selection);

   // get current instrument
   unsigned char Instr[INSTRLEN];
   Global::VoiceArray[Selection+1].GetInstr(Instr);
   PopulateInstr(Instr);
   m_CurrentSelection = Selection;
}

void CInstrEdit::OnIeRename() 
{
   CInstrRename dlgInstrRename;

   CString s;
   CComboBox * cboPresets = (CComboBox *) GetDlgItem(IDC_IE_PRESET_LIST);
   cboPresets->GetWindowText(s);   
   // strip the voice number
   dlgInstrRename.SetRenameValue(s.Right(s.GetLength() - 3));

   if (dlgInstrRename.DoModal() == IDOK)
   {
      int Pos = cboPresets->GetCurSel();
      CString NewName;
      NewName.Format("%02d %s", Pos+1, dlgInstrRename.GetRenameValue());

      cboPresets->DeleteString(Pos);
      cboPresets->InsertString(Pos, NewName);
      cboPresets->SetCurSel(Pos);
      Global::VoiceArray[Pos+1].SetName(NewName);
      Global::pMainFrm->ToolBar_FillCombo();
   }
}

void CInstrEdit::OnIeLoad() 
{
   CFile fWorkspace;
   char NullChar = 0;
   char CurDir[MAX_PATH+1];

   CFileDialog dlgFileOpen(true, "INS", NULL, OFN_HIDEREADONLY | OFN_PATHMUSTEXIST | OFN_FILEMUSTEXIST,
      szInstrLoadFilter, NULL);
   GetCurrentDirectory(MAX_PATH, CurDir);
   dlgFileOpen.m_ofn.lpstrInitialDir = CurDir;
   if (dlgFileOpen.DoModal() == IDCANCEL)
      return;
	
   if (!fWorkspace.Open(dlgFileOpen.GetPathName(), CFile::modeRead))
   {
      CString ErrMsg;
      ErrMsg = "Failed to Open " + dlgFileOpen.GetFileName();
      AfxMessageBox(ErrMsg);
      return;
   }
      
	// MVSTracker Instruments
	if( stricmp(strrchr(dlgFileOpen.GetFileName(), '.'), ".ins") == 0)
	{   
		// validate header
		char Header[sizeof(szInstrHeader)];
		fWorkspace.Read(Header, sizeof(szInstrHeader)-1);
		if (memcmp(Header, szInstrHeader, sizeof(szInstrHeader)-1) != 0)
		{
		  AfxMessageBox("Invalid MVS Tracker Instrument file");
		  return;
		}

		CComboBox * cboPresets = (CComboBox *) GetDlgItem(IDC_IE_PRESET_LIST);
		int Selection = cboPresets->GetCurSel();   

		// get voice name and data
		CString RawInstrName = Global::ReadNullTerminatedString(&fWorkspace);
		CString InstrName;
		InstrName.Format("%02d %s", Selection+1, RawInstrName);
		unsigned char Instr[INSTRLEN];

		fWorkspace.Read(Instr, INSTRLEN);

		// store
		Global::VoiceArray[Selection+1].SetName(InstrName);
		Global::VoiceArray[Selection+1].SetInstr(Instr);

		fWorkspace.Close();

		// update the dialog box
		InitialiseCombo();
		UpdateSelection(Selection);

		// update main window's combo box 
		Global::pMainFrm->ToolBar_FillCombo();

	}
	else
	{
		// Gens Kmod Channel Dump
		if( stricmp(strrchr(dlgFileOpen.GetFileName(), '.'), ".y12") == 0)
		{
			CComboBox * cboPresets = (CComboBox *) GetDlgItem(IDC_IE_PRESET_LIST);
			int Selection = cboPresets->GetCurSel();   

			// get voice name and data
			CString RawInstrName = dlgFileOpen.GetFileTitle();
			CString InstrName;
			InstrName.Format("%02d %s", Selection+1, RawInstrName);
			unsigned char Instr[INSTRLEN];
			unsigned char Instr_KMod[128];

			fWorkspace.Read(Instr_KMod, 128);

			// convert kmod file dump
			for(int i=0;i<6;i++)
			{	Instr[(i*4)+0]=Instr_KMod[0+i];
				Instr[(i*4)+1]=Instr_KMod[16+i];
				Instr[(i*4)+2]=Instr_KMod[32+i];
				Instr[(i*4)+3]=Instr_KMod[48+i];
			}
			unsigned char val2=Instr_KMod[65];
			unsigned char val1=Instr_KMod[64];
			Instr[24]=((val2 & 7) << 3) | (val1 & 7);
			
			// store
			Global::VoiceArray[Selection+1].SetName(InstrName);
			Global::VoiceArray[Selection+1].SetInstr(Instr);

			fWorkspace.Close();

			// update the dialog box
			InitialiseCombo();
			UpdateSelection(Selection);

			// update main window's combo box 
			Global::pMainFrm->ToolBar_FillCombo();
			
		}
		else
		{
			AfxMessageBox("Invalid MVS Tracker Instrument file");
			return;
		}
	}

}


void CInstrEdit::OnIeSave() 
{
   // first ensure that the current settings are stored
   OnIeStore();

   CFile fWorkspace;
   char NullChar = 0;
   char CurDir[MAX_PATH+1];
   char Filename[MAX_PATH+1];
   CComboBox * cboPresets = (CComboBox *) GetDlgItem(IDC_IE_PRESET_LIST);
   int Selection = cboPresets->GetCurSel();   

   // convert CString to strip off instr# and convert to char array
   CString VoiceName = Global::VoiceArray[Selection+1].GetName();
   VoiceName = VoiceName.Right(VoiceName.GetLength() - 3);
   memcpy(Filename, VoiceName.GetBuffer(MAX_PATH), VoiceName.GetLength());
   Filename[VoiceName.GetLength()] = '\0';

   CFileDialog dlgFileSave(false, "INS", NULL, OFN_HIDEREADONLY | OFN_OVERWRITEPROMPT,
      szInstrFilter, NULL);
   GetCurrentDirectory(MAX_PATH, CurDir);
   dlgFileSave.m_ofn.lpstrInitialDir = CurDir;
   dlgFileSave.m_ofn.lpstrFile = Filename;
   dlgFileSave.m_ofn.nMaxFile = MAX_PATH;
   if (dlgFileSave.DoModal() == IDCANCEL)
      return;
   
   if (!fWorkspace.Open(dlgFileSave.GetPathName(), CFile::modeWrite | CFile::modeCreate))
   {
      CString ErrMsg;
      ErrMsg = "Failed to Create " + dlgFileSave.GetFileName();
      AfxMessageBox(ErrMsg);
      return;
   }

   // write header 'MVSI' + ver num of save format
   CString Header(szInstrHeader);
   fWorkspace.Write(Header, sizeof(szInstrHeader)-1);

   // save instrument
   unsigned char Instr[INSTRLEN];
   fWorkspace.Write(VoiceName, VoiceName.GetLength());
   fWorkspace.Write(&NullChar, 1); // null terminate string
   Global::VoiceArray[Selection+1].GetInstr(Instr);
   fWorkspace.Write(Instr, INSTRLEN);

   fWorkspace.Close();
}


// user wishes to save the instrument 
void CInstrEdit::OnIeStore() 
{
   CComboBox * cboPresets = (CComboBox *) GetDlgItem(IDC_IE_PRESET_LIST);
   int Selection = cboPresets->GetCurSel();   

   unsigned char Instr[INSTRLEN];
   BuildInstrArray(Instr);
   
   Global::VoiceArray[Selection+1].SetInstr(Instr);
}

void CInstrEdit::OnSelchangeIePresetList() 
{
   CComboBox * cboPresets = (CComboBox *) GetDlgItem(IDC_IE_PRESET_LIST);
   int Selection = cboPresets->GetCurSel();   

   // before we populate the on screen params with the new selection
   // best check if there are changes that may need saving
   if (Selection != m_CurrentSelection)
   {
      // get the old one
      unsigned char OldInstr[INSTRLEN];
      Global::VoiceArray[m_CurrentSelection+1].GetInstr(OldInstr);

      // build one from the current params
      BuildInstrArray(m_CurrentInstr);
      
      // compare
      if (memcmp(OldInstr, m_CurrentInstr, INSTRLEN) != 0)
      {
         CString Msg;
         Msg.Format("Store changes to Instrument %d first?", m_CurrentSelection+1);
         if (AfxMessageBox(Msg, MB_YESNO) == IDYES)
            Global::VoiceArray[m_CurrentSelection+1].SetInstr(m_CurrentInstr);
      }

   }

   m_CurrentSelection = Selection;
   // get current instrument
   unsigned char Instr[INSTRLEN];
   Global::VoiceArray[Selection+1].GetInstr(Instr);
   PopulateInstr(Instr);
}

void CInstrEdit::OnIeClear() 
{
   // fill the display with 0
   unsigned char Instr[INSTRLEN];
   memset(Instr, 0, INSTRLEN);
   PopulateInstr(Instr);
}

void CInstrEdit::OnShowWindow(BOOL bShow, UINT nStatus) 
{
	CDialog::OnShowWindow(bShow, nStatus);

   InitialiseCombo();
}
