// MainFrm.cpp : implementation of the CMainFrame class
//

#include "stdafx.h"
#include "MVSTracker.h"

#include "MainFrm.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

/////////////////////////////////////////////////////////////////////////////
// CMainFrame

IMPLEMENT_DYNAMIC(CMainFrame, CFrameWnd)

BEGIN_MESSAGE_MAP(CMainFrame, CFrameWnd)
	//{{AFX_MSG_MAP(CMainFrame)
	ON_WM_CREATE()
	ON_WM_SETFOCUS()
	ON_COMMAND(ID_VIEW_INSTR_EDIT, OnViewInstrEdit)
	ON_BN_CLICKED(IDC_PL_PAT_NEW, OnPlPatNew)
	ON_BN_CLICKED(IDC_PH_BUT_EDIT, OnPhButEdit)
	ON_BN_CLICKED(IDC_PL_PAT_DOWN, OnPlPatDown)
	ON_BN_CLICKED(IDC_PL_PAT_UP, OnPlPatUp)
	ON_BN_CLICKED(IDC_PL_PAT_RENAME, OnPlPatRename)
	ON_BN_CLICKED(IDC_PL_SONG_NEW, OnPlSongNew)
	ON_BN_CLICKED(IDC_PL_SONG_INSERT, OnPlSongInsert)
	ON_BN_CLICKED(IDC_PL_SONG_UPDATE, OnPlSongUpdate)
	ON_BN_CLICKED(IDC_PL_SONG_DELETE, OnPlSongDelete)
	ON_LBN_SELCHANGE(IDC_PL_LIST, OnSelchangePlList)
	ON_COMMAND(ID_EDIT_COPYPATTERN, OnEditCopypattern)
	ON_COMMAND(ID_EDIT_COPYTRACK, OnEditCopytrack)
	ON_COMMAND(ID_EDIT_CUTTRACK, OnEditCuttrack)
	ON_COMMAND(ID_EDIT_PASTEPATTERN, OnEditPastepattern)
	ON_COMMAND(ID_EDIT_PASTETRACK, OnEditPastetrack)
	ON_COMMAND(ID_VIEW_ROMMAKER, OnViewRommaker)
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

static UINT indicators[] =
{
	ID_SEPARATOR,           // status line indicator
	ID_INDICATOR_CAPS,
	ID_INDICATOR_NUM,
	ID_INDICATOR_SCRL,
};

/////////////////////////////////////////////////////////////////////////////
// CMainFrame construction/destruction

CMainFrame::CMainFrame()
{
   Global::pMainFrm = this;
}

CMainFrame::~CMainFrame()
{
   if (Global::pInstrEdit != NULL)
      delete Global::pInstrEdit;

   if (Global::pROMMaker != NULL)
      delete Global::pROMMaker;
}

int CMainFrame::OnCreate(LPCREATESTRUCT lpCreateStruct)
{
	if (CFrameWnd::OnCreate(lpCreateStruct) == -1)
		return -1;
	// create a view to occupy the client area of the frame
	if (!m_wndView.Create(NULL, NULL, AFX_WS_DEFAULT_VIEW,
		CRect(0, 0, 0, 0), this, AFX_IDW_PANE_FIRST, NULL))
	{
		TRACE0("Failed to create view window\n");
		return -1;
	}

	if (!m_wndStatusBar.Create(this) ||
		!m_wndStatusBar.SetIndicators(indicators,
		  sizeof(indicators)/sizeof(UINT)))
	{
		TRACE0("Failed to create status bar\n");
		return -1;      // fail to create
	}

   if (!m_wndPatEdToolBar.Create(this, IDD_PATTERN_HEADER, CBRS_TOP, 0))
	{
		TRACE0("Failed to create pattern editor bar\n");
		return -1;      // fail to create
	}

   // initialise the toolbar
   m_bmpPlay.LoadBitmap(IDB_PH_BUT_PLAY);
   CButton * cb = (CButton *)m_wndPatEdToolBar.GetDlgItem(IDC_PH_BUT_PLAY);
   cb->SetBitmap(m_bmpPlay);

   m_bmpStop.LoadBitmap(IDB_PH_BUT_STOP);
   cb = (CButton *)m_wndPatEdToolBar.GetDlgItem(IDC_PH_BUT_STOP);
   cb->SetBitmap(m_bmpStop);

   m_bmpPlayAll.LoadBitmap(IDB_PH_BUT_PLAY_ALL);
   cb = (CButton *)m_wndPatEdToolBar.GetDlgItem(IDC_PH_BUT_PLAY_ALL);
   cb->SetBitmap(m_bmpPlayAll);

   if (!m_wndPatListToolBar.Create(this, IDD_PATTERN_LIST, CBRS_LEFT, 0))
	{
		TRACE0("Failed to create pattern list bar\n");
		return -1;      // fail to create
	}

   // NOTE the toolbars are initialised in the main class as they have
   // to be done after the globals are initialised


   // initialise view dialogs
   Global::pInstrEdit = new CInstrEdit(this);

#ifndef RELEASE_CRIPPLED
   Global::pROMMaker = new CROMMaker(this);
#endif

   // initialise Thor's MVS Sound DLL
   initSound(m_hWnd, 44100);

	return 0;
}

BOOL CMainFrame::PreCreateWindow(CREATESTRUCT& cs)
{
	if( !CFrameWnd::PreCreateWindow(cs) )
		return FALSE;
	cs.dwExStyle &= ~WS_EX_CLIENTEDGE;
	cs.lpszClass = AfxRegisterWndClass(0);
	return TRUE;
}

/////////////////////////////////////////////////////////////////////////////
// CMainFrame diagnostics

#ifdef _DEBUG
void CMainFrame::AssertValid() const
{
	CFrameWnd::AssertValid();
}

void CMainFrame::Dump(CDumpContext& dc) const
{
	CFrameWnd::Dump(dc);
}

#endif //_DEBUG

/////////////////////////////////////////////////////////////////////////////
// CMainFrame message handlers
void CMainFrame::OnSetFocus(CWnd* pOldWnd)
{
	// forward focus to the view window
	m_wndView.SetFocus();
}

BOOL CMainFrame::OnCmdMsg(UINT nID, int nCode, void* pExtra, AFX_CMDHANDLERINFO* pHandlerInfo)
{
	// let the view have first crack at the command
	if (m_wndView.OnCmdMsg(nID, nCode, pExtra, pHandlerInfo))
		return TRUE;

	// otherwise, do default handling
	return CFrameWnd::OnCmdMsg(nID, nCode, pExtra, pHandlerInfo);
}

/////////////////////////////////////////////////////////////////////////////
// Menu Handlers
void CMainFrame::OnViewInstrEdit() 
{
   CMenu* pMenu = GetMenu();
   if(pMenu->GetMenuState(ID_VIEW_INSTR_EDIT, MF_UNCHECKED) == MF_UNCHECKED)
   {
      Global::pInstrEdit->ShowWindow(SW_SHOWNORMAL);   
      pMenu->CheckMenuItem(ID_VIEW_INSTR_EDIT, MF_CHECKED | MF_BYCOMMAND);
      // get the instrument editor to display the current instrument
      CComboBox * cc = (CComboBox *) m_wndPatEdToolBar.GetDlgItem(IDC_PH_VOICES);
      Global::pInstrEdit->UpdateSelection(cc->GetCurSel());
   }
   else
   {
      Global::pInstrEdit->ShowWindow(SW_HIDE);   
      pMenu->CheckMenuItem(ID_VIEW_INSTR_EDIT, MF_UNCHECKED | MF_BYCOMMAND);
   }
}

void CMainFrame::OnViewRommaker() 
{
#ifndef RELEASE_CRIPPLED

   CMenu* pMenu = GetMenu();
   if(pMenu->GetMenuState(ID_VIEW_ROMMAKER, MF_UNCHECKED) == MF_UNCHECKED)
   {
      Global::pROMMaker->ShowWindow(SW_SHOWNORMAL);   
      pMenu->CheckMenuItem(ID_VIEW_ROMMAKER, MF_CHECKED | MF_BYCOMMAND);
   }
   else
   {
      Global::pROMMaker->ShowWindow(SW_HIDE);   
      pMenu->CheckMenuItem(ID_VIEW_ROMMAKER, MF_UNCHECKED | MF_BYCOMMAND);
   }
#endif
}





// invoke InstrEditor
void CMainFrame::OnPhButEdit() 
{
   Global::pInstrEdit->ShowWindow(SW_SHOWNORMAL);   
   CMenu* pMenu = GetMenu();
   pMenu->CheckMenuItem(ID_VIEW_INSTR_EDIT, MF_CHECKED | MF_BYCOMMAND);

   // get the instrument editor to display the current instrument
   CComboBox * cc = (CComboBox *) m_wndPatEdToolBar.GetDlgItem(IDC_PH_VOICES);
   Global::pInstrEdit->UpdateSelection(cc->GetCurSel());
}



/////////////////////////////////////////////////////////////////////////////
// Song Editor




/////////////////////////////////////////////////////////////////////////////
// Tool Bar helper functions
// accessed via Global::pMainFrm->ToolBar_...()
//
void CMainFrame::InitialiseToolBars()
{
   // initialise the pattern header toolbar
   ToolBar_SetOctave(4);
   CSpinButtonCtrl * cs = (CSpinButtonCtrl *)m_wndPatEdToolBar.GetDlgItem(IDC_PH_OCT_SPIN);
   cs->SetRange(LOWER_OCT_RANGE, UPPER_OCT_RANGE);
   ToolBar_FillCombo();

   // initialise the pattern list toolbar
   CListBox * cl = (CListBox *)m_wndPatListToolBar.GetDlgItem(IDC_PL_LIST);
   cl->ResetContent();

   OnPlPatNew(); // default pattern   
   cl->AddString("00 - 00");
   cl->SetCurSel(0);
}

void CMainFrame::ToolBar_SetOctave(int Value)
{
   m_wndPatEdToolBar.SetDlgItemInt(IDC_PH_OCTAVE, Value);
}

int CMainFrame::ToolBar_GetOctave()
{
   return m_wndPatEdToolBar.GetDlgItemInt(IDC_PH_OCTAVE);
}

void CMainFrame::ToolBar_SetCurrentVoice(int Value)
{
   CComboBox * cc = (CComboBox *) m_wndPatEdToolBar.GetDlgItem(IDC_PH_VOICES);
   cc->SetCurSel(Value);
}

int CMainFrame::ToolBar_GetCurrentVoice()
{
   CComboBox * cc = (CComboBox *) m_wndPatEdToolBar.GetDlgItem(IDC_PH_VOICES);
   return cc->GetCurSel();
}

void CMainFrame::ToolBar_FillCombo()
{
   CComboBox * cc = (CComboBox *) m_wndPatEdToolBar.GetDlgItem(IDC_PH_VOICES);

   int Current = cc->GetCurSel();
   if (Current == -1) // no selection
      Current = 0;

   cc->ResetContent();
   for (int i = 1; i < MAX_VOICES; ++i)
      cc->AddString(Global::VoiceArray[i].GetName());

   cc->SetCurSel(Current);
}

int CMainFrame::ToolBar_GetPattern()
{
   return m_wndPatListToolBar.GetDlgItemInt(IDC_PL_PAT_NUM);
}


int CMainFrame::ToolBar_GetSongLen()
{
	CListBox * cl = (CListBox *) m_wndPatListToolBar.GetDlgItem(IDC_PL_LIST);
   return cl->GetCount();
}

int CMainFrame::ToolBar_GetSongPattern(int Pos)
{
	CListBox * cl = (CListBox *) m_wndPatListToolBar.GetDlgItem(IDC_PL_LIST);

   CString CurText;
   cl->GetText(Pos, CurText);
   
   return _ttoi((LPCTSTR)CurText.Right(2));
}

void CMainFrame::ToolBar_AddSongPattern(int PatNum)
{
	CListBox * cl = (CListBox *) m_wndPatListToolBar.GetDlgItem(IDC_PL_LIST);

   int Count = cl->GetCount();

   CString sPatEntry;
   sPatEntry.Format("%02d - %02d", Count, PatNum);
   cl->AddString(sPatEntry);
   cl->SetCurSel(0);
}


void CMainFrame::ToolBar_ClearSongPatterns()
{
   // initialise the pattern list toolbar
   CListBox * cl = (CListBox *)m_wndPatListToolBar.GetDlgItem(IDC_PL_LIST);
   cl->ResetContent();
}

void CMainFrame::ToolBar_RefreshPatName()
{
   // get current pattern number
   int PatNum = m_wndPatListToolBar.GetDlgItemInt(IDC_PL_PAT_NUM);

   CString PatName = Global::Patterns[PatNum].GetPatternName();
   m_wndPatListToolBar.SetDlgItemText(IDC_PL_PAT_NAME, PatName);

   Invalidate(); // cause new pattern redraw
}



void CMainFrame::OnPlPatNew() 
{
   if (Global::NumPatterns < MAX_PATTERNS)
   {
      // add new pattern
      Global::NumPatterns++;
      CPattern * pPattern = &Global::Patterns[Global::NumPatterns-1];
      pPattern->Clear();

      CString PatNum;
      PatNum.Format("%02d", Global::NumPatterns-1);
      m_wndPatListToolBar.SetDlgItemText(IDC_PL_PAT_NUM, PatNum);

      CString PatName;
      PatName.Format("Pat %02d", Global::NumPatterns-1);
      m_wndPatListToolBar.SetDlgItemText(IDC_PL_PAT_NAME, PatName);
      Global::Patterns[Global::NumPatterns-1].SetPatternName(PatName);

      Invalidate(); // cause new pattern redraw
   }
}



void CMainFrame::OnPlPatDown() 
{
   // get current pattern number
   int PatNum = m_wndPatListToolBar.GetDlgItemInt(IDC_PL_PAT_NUM);
   
   // decrement the pattern number
   if (PatNum > 0)
   {
      PatNum --;
      CString sPatNum;
      sPatNum.Format("%02d", PatNum);
      m_wndPatListToolBar.SetDlgItemText(IDC_PL_PAT_NUM, sPatNum);

      CString PatName = Global::Patterns[PatNum].GetPatternName();
      m_wndPatListToolBar.SetDlgItemText(IDC_PL_PAT_NAME, PatName);

      Invalidate(); // cause new pattern redraw
   }
}


void CMainFrame::OnPlPatUp() 
{
   // get current pattern number
   int PatNum = m_wndPatListToolBar.GetDlgItemInt(IDC_PL_PAT_NUM);
   
   // increase the pattern number
   if (PatNum < Global::NumPatterns - 1)
   {
      PatNum ++;
      CString sPatNum;
      sPatNum.Format("%02d", PatNum);
      m_wndPatListToolBar.SetDlgItemText(IDC_PL_PAT_NUM, sPatNum);

      CString PatName = Global::Patterns[PatNum].GetPatternName();
      m_wndPatListToolBar.SetDlgItemText(IDC_PL_PAT_NAME, PatName);

      Invalidate(); // cause new pattern redraw
   }
}

void CMainFrame::OnPlPatRename() 
{
   // get current pattern number
   int PatNum = m_wndPatListToolBar.GetDlgItemInt(IDC_PL_PAT_NUM);

   CString PatName;
   m_wndPatListToolBar.GetDlgItemText(IDC_PL_PAT_NAME, PatName);
   Global::Patterns[PatNum].SetPatternName(PatName);
}

void CMainFrame::OnPlSongNew() 
{
	CListBox * cl = (CListBox *) m_wndPatListToolBar.GetDlgItem(IDC_PL_LIST);
   int Count = cl->GetCount();

   int PatNum = m_wndPatListToolBar.GetDlgItemInt(IDC_PL_PAT_NUM);

   CString sPatEntry;
   sPatEntry.Format("%02d - %02d", Count, PatNum);
   cl->AddString(sPatEntry);
   cl->SetCurSel(Count);
}

void CMainFrame::OnPlSongInsert() 
{
	CListBox * cl = (CListBox *) m_wndPatListToolBar.GetDlgItem(IDC_PL_LIST);
   int Sel = cl->GetCurSel();
   int PatNum = m_wndPatListToolBar.GetDlgItemInt(IDC_PL_PAT_NUM);

   CString sPatEntry;
   sPatEntry.Format("%02d - %02d", Sel, PatNum);
   cl->InsertString(Sel, sPatEntry);
   RenumberSongList();
   cl->SetCurSel(Sel);
}

void CMainFrame::OnPlSongUpdate() 
{
	CListBox * cl = (CListBox *) m_wndPatListToolBar.GetDlgItem(IDC_PL_LIST);
   int Sel = cl->GetCurSel();
   int PatNum = m_wndPatListToolBar.GetDlgItemInt(IDC_PL_PAT_NUM);

   CString sPatEntry;
   sPatEntry.Format("%02d - %02d", Sel, PatNum);
   cl->DeleteString(Sel);
   cl->InsertString(Sel, sPatEntry);
   cl->SetCurSel(Sel);
}

void CMainFrame::OnPlSongDelete() 
{
	CListBox * cl = (CListBox *) m_wndPatListToolBar.GetDlgItem(IDC_PL_LIST);
   int Count = cl->GetCount();
   int Sel = cl->GetCurSel();

   // delete current entry unless it is the last
   if (Count > 1)
   {
      cl->DeleteString(Sel);
      if (Sel > 0)
         Sel--;
   }

   RenumberSongList();
   cl->SetCurSel(Sel);
}


// Rebuild the song number to remove any gaps so that the "nn - xx" 
// nn is always sequential 
void CMainFrame::RenumberSongList()
{
	CListBox * cl = (CListBox *) m_wndPatListToolBar.GetDlgItem(IDC_PL_LIST);

   for (int i = 0; i <  cl->GetCount(); i ++)
   {
      CString CurText;
      cl->GetText(i, CurText);
      cl->DeleteString(i);
      CString NewText;
      NewText.Format("%02d - %s", i, CurText.Right(2));
      cl->InsertString(i, NewText);
   }
}

void CMainFrame::OnSelchangePlList() 
{
	CListBox * cl = (CListBox *) m_wndPatListToolBar.GetDlgItem(IDC_PL_LIST);
   
   // get pattern number from song
   int Sel = cl->GetCurSel();
   CString SongRow;
   cl->GetText(Sel, SongRow);
   int PatNum = atoi(SongRow.Right(2));

   CString sPatNum;
   sPatNum.Format("%02d", PatNum);
   m_wndPatListToolBar.SetDlgItemText(IDC_PL_PAT_NUM, sPatNum);

   CString PatName = Global::Patterns[PatNum].GetPatternName();
   m_wndPatListToolBar.SetDlgItemText(IDC_PL_PAT_NAME, PatName);

   Invalidate(); // cause new pattern redraw
}



// --------------------------------------------------------------------------
//
//                              CLIPBOARD HANDLING 
//
// --------------------------------------------------------------------------


void CMainFrame::CopyTrack(int PatNum, int ChanNum, int ClipChan)
{
   // copy specified track to clipboard
   for (int i = 0; i < PATSIZE; i++)
   {
      Global::Clipboard[ClipChan][i].Note = Global::Patterns[PatNum].GetNote(ChanNum, i);
      Global::Clipboard[ClipChan][i].Oct = Global::Patterns[PatNum].GetOct(ChanNum, i);
      Global::Clipboard[ClipChan][i].Voice = Global::Patterns[PatNum].GetVoice(ChanNum, i);
      Global::Clipboard[ClipChan][i].Effect = Global::Patterns[PatNum].GetEffect(ChanNum, i);
      Global::Clipboard[ClipChan][i].Value = Global::Patterns[PatNum].GetValue(ChanNum, i);
   }
}

void CMainFrame::PasteTrack(int PatNum, int ChanNum, int ClipChan)
{
   // paste specified track from clipboard
   for (int i = 0; i < PATSIZE; i++)
   {
      Global::Patterns[PatNum].SetNote(ChanNum, i, Global::Clipboard[ClipChan][i].Note);
      Global::Patterns[PatNum].SetOct(ChanNum, i, Global::Clipboard[ClipChan][i].Oct);
      Global::Patterns[PatNum].SetVoice(ChanNum, i, Global::Clipboard[ClipChan][i].Voice);
      Global::Patterns[PatNum].SetEffect(ChanNum, i, Global::Clipboard[ClipChan][i].Effect);
      Global::Patterns[PatNum].SetValue(ChanNum, i, Global::Clipboard[ClipChan][i].Value);
   }
}

void CMainFrame::OnEditCopytrack() 
{
   int ChanNum = m_wndView.GetCurrentXPos() / MAXITEMS;
   int PatNum = ToolBar_GetPattern();
   CopyTrack(PatNum, ChanNum, 0);
}

void CMainFrame::OnEditCuttrack() 
{
   // first copy track and then clear current track
   int ChanNum = m_wndView.GetCurrentXPos() / MAXITEMS;
   int PatNum = ToolBar_GetPattern();
   CopyTrack(PatNum, ChanNum, 0);

   // clear current track
   for (int i = 0; i < PATSIZE; i++)
   {
      Global::Patterns[PatNum].SetNote(ChanNum, i, NOTE_BLANK);
      Global::Patterns[PatNum].SetOct(ChanNum, i, -1);
      Global::Patterns[PatNum].SetVoice(ChanNum, i, -1);
      Global::Patterns[PatNum].SetEffect(ChanNum, i, -1);
      Global::Patterns[PatNum].SetValue(ChanNum, i, -1);
   }

   // force pattern redraw
   Invalidate();
}
void CMainFrame::OnEditPastetrack() 
{
   int ChanNum = m_wndView.GetCurrentXPos() / MAXITEMS;
   int PatNum = ToolBar_GetPattern();
   PasteTrack(PatNum, ChanNum, 0);

   // force pattern redraw
   Invalidate();
}


void CMainFrame::OnEditCopypattern() 
{
   int PatNum = ToolBar_GetPattern();
   for (int i = 0; i < MAX_CHANNELS; i++)
      CopyTrack(PatNum, i, i);
}

void CMainFrame::OnEditPastepattern() 
{
   int PatNum = ToolBar_GetPattern();
   for (int i = 0; i < MAX_CHANNELS; i++)
      PasteTrack(PatNum, i, i);

   // force pattern redraw
   Invalidate();
}



