/////////////////////////////////////////////////////////////////////////////
// CPattern implementation



#include "stdafx.h"



void CPattern::Clear()
{
   for (int c=0; c < MAX_CHANNELS; c++)
   {
      for (int r=0; r<PATSIZE; r++)
      {
         mChannels[c][r].Note = NOTE_BLANK;
         mChannels[c][r].Oct = -1;
         mChannels[c][r].Voice = -1;
         mChannels[c][r].Effect = -1;
         mChannels[c][r].Value = -1;
      }
   }
}


unsigned char * CPattern::BuildPattern(bool bUseActiveChannels)
{
   int bufPtr = 0;

   for (int row = 0; row < PATSIZE; row ++)
   {
      for (int channel = 0; channel < MAX_CHANNELS; channel++)
      {
         // byte1 = oct | note
         int Note = GetNote(channel, row);
         
         if (bUseActiveChannels && !Global::bChannelActive[channel])
            Note = NOTE_BLANK;

         int ConvNote = 0;
         // convert between note formats
         switch (Note)
         {
         case 0: // c
            ConvNote = 12;
            break;
         case NOTE_BLANK:
            ConvNote = 0;
            break;
         default:
            ConvNote = Note;
            break;
         }


         int Oct = GetOct(channel, row);
         if (Oct == -1)
         {
            mBuiltPattern[bufPtr++] = 0;
         }
         else
         {
            if (ConvNote == 12) Oct --;
            mBuiltPattern[bufPtr++] = ConvNote | (Oct << 4);
         }
      
         // instrument
         mBuiltPattern[bufPtr++] = GetVoice(channel, row);

         // effect/value
         mBuiltPattern[bufPtr++] = GetEffect(channel, row);
         mBuiltPattern[bufPtr++] = GetValue(channel, row);
      }
   }

   return mBuiltPattern;
}

void CPattern::SetPattern(unsigned char * pNewPat)
{
   int bufPtr = 0;

   for (int row = 0; row < PATSIZE; row ++)
   {
      for (int channel = 0; channel < MAX_CHANNELS; channel++)
      {
         int Oct = pNewPat[bufPtr] >> 4;
         int Note = pNewPat[bufPtr++] & 15;

         int ConvNote = 0;

         // convert between note formats
         switch (Note)
         {
         case 12: // c
            ConvNote = 0;
            break;
         case 0:
            ConvNote = NOTE_BLANK;
            break;
         default:
            ConvNote = Note;
            break;
         }
         
         SetNote(channel, row, ConvNote);
         
         if (Oct == 0)
         {
            SetOct(channel, row, -1);
         }
         else
         {
            if (ConvNote == 0) Oct ++;
            SetOct(channel, row, Oct);
         }
      
         // instrument
         SetVoice(channel, row, pNewPat[bufPtr++]);

         // effect/value
         SetEffect(channel, row, pNewPat[bufPtr++]);
         SetValue(channel, row, pNewPat[bufPtr++]);
      }
   }
}


// returns the largest instrument number used in this pattern
int CPattern::GetHighestInstrNo()
{
   int HighestInstrNo = 1;

   for (int row = 0; row < PATSIZE; row ++)
   {
      for (int channel = 0; channel < MAX_CHANNELS; channel++)
      {
         int InstrNo = GetVoice(channel, row);
         if (InstrNo > HighestInstrNo)
            HighestInstrNo = InstrNo;
      }
   }
   return HighestInstrNo;
}
