// ROMMaker.cpp : implementation file
//

#include "stdafx.h"
#include "MVSTracker.h"
#include "ROMMaker.h"
#include "ROMMakerHelper.h"
#include "modplay.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

/////////////////////////////////////////////////////////////////////////////
// CROMMaker dialog


CROMMaker::CROMMaker(CWnd* pParent /*=NULL*/)
	: CDialog(CROMMaker::IDD, pParent)
{
	//{{AFX_DATA_INIT(CROMMaker)
		// NOTE: the ClassWizard will add member initialization here
	//}}AFX_DATA_INIT

   Create(IDD_ROM_MAKER);
   mpParent = pParent;
}


CROMMaker::~CROMMaker()
{
   while (m_MusModules.GetCount() > 0)
   {
      CROMMakerHelper * pRMH = m_MusModules.GetHead();
      delete pRMH;
      m_MusModules.RemoveHead();
   }

   while (m_SampModules.GetCount() > 0)
   {
      CSampleEncoding * pSE = m_SampModules.GetHead();
      delete pSE;
      m_SampModules.RemoveHead();
   }
}


BOOL CROMMaker::OnInitDialog() 
{
	CDialog::OnInitDialog();
	
   TCHAR  szWorkDir[MAX_PATH];
    
   // Here we subclass our CDirTreeCtrl
   if ( !m_TreeCtrl.m_hWnd )
   {
      if ( m_TreeCtrl.SubclassDlgItem( IDC_RM_DIRTREE, this ) )
      {           
         m_TreeCtrl.DisplayTree( NULL /*Display all*/, 
                                 TRUE /* TRUE = Display Files*/ );

         GetCurrentDirectory(MAX_PATH, szWorkDir);
         // set the Path to the current Work-Directory
         m_TreeCtrl.SetSelPath( szWorkDir );
      }
   }

   UpdateSpaceRemaining();
	
	return TRUE;  // return TRUE unless you set the focus to a control
	              // EXCEPTION: OCX Property Pages should return FALSE
}


void CROMMaker::DoDataExchange(CDataExchange* pDX)
{
	CDialog::DoDataExchange(pDX);
	//{{AFX_DATA_MAP(CROMMaker)
		// NOTE: the ClassWizard will add DDX and DDV calls here
	//}}AFX_DATA_MAP
}


BEGIN_MESSAGE_MAP(CROMMaker, CDialog)
	//{{AFX_MSG_MAP(CROMMaker)
	ON_BN_CLICKED(IDC_RM_ADD, OnRmAdd)
	ON_BN_CLICKED(IDC_RM_REMOVE, OnRmRemove)
	ON_BN_CLICKED(IDC_RM_MAKEROMS, OnRmMakeroms)
	ON_LBN_SELCHANGE(IDC_RM_MUSICFILES, OnSelchangeRmMusicfiles)
	ON_LBN_SELCHANGE(IDC_RM_SAMPLEFILES, OnSelchangeRmSamplefiles)
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// ROMMaker message handlers

void CROMMaker::OnCancel() 
{
   ShowWindow(SW_HIDE);
   
   CMenu* pMenu = mpParent->GetMenu();
   pMenu->CheckMenuItem(ID_VIEW_ROMMAKER, MF_UNCHECKED | MF_BYCOMMAND);
	
	CDialog::OnCancel();
}

void CROMMaker::OnRmAdd() 
{
	HTREEITEM Item = m_TreeCtrl.GetSelectedItem();

   // ensure its a file
   CString PathName = m_TreeCtrl.GetFullPath(Item);
   CFile theFile;
   CFileStatus fileStatus;
   theFile.GetStatus(PathName, fileStatus);
   if (fileStatus.m_attribute & 0x10)
   {
      AfxMessageBox("You Must Select a MUS or WAV File");
      return;
   }

   // add name to list box
   CString FileName = m_TreeCtrl.GetItemText(Item);

   // is it a MUS or a WAV?
   if (FileName.Right(3).CompareNoCase("MUS") == 0)
   {
      // Adding a MUS
      CListBox * cl = (CListBox *) GetDlgItem(IDC_RM_MUSICFILES);
      cl->AddString(FileName);
      CROMMakerHelper * pRMH = new CROMMakerHelper;
      m_MusModules.AddTail(pRMH);
      int Index = cl->GetCount()-1;

         // check for success, if not tidy up
      if (!pRMH->LoadMusic(PathName, FileName))
      {
         delete pRMH;
         cl->DeleteString(Index);
         m_MusModules.RemoveTail();
      }
   }
   else
   {
      // Adding a WAV
      CListBox * cl = (CListBox *) GetDlgItem(IDC_RM_SAMPLEFILES);
      cl->AddString(FileName);
      CSampleEncoding * pSE = new CSampleEncoding;
      m_SampModules.AddTail(pSE);
      int Index = cl->GetCount() -1;

      // check for success, if not tidy up
      if (!pSE->ValidateFile(PathName, FileName))
      {
         delete pSE;
         cl->DeleteString(Index);
         m_SampModules.RemoveTail();
      }
   }

   UpdateSpaceRemaining();
}

void CROMMaker::OnRmRemove() 
{
   CListBox * clMusic = (CListBox *) GetDlgItem(IDC_RM_MUSICFILES);
   CListBox * clSamples = (CListBox *) GetDlgItem(IDC_RM_SAMPLEFILES);

   if (clMusic->GetCurSel() == LB_ERR && clSamples->GetCurSel() == LB_ERR)
   {
      AfxMessageBox("You Must Select Either a Music or Sample File to Remove");
      return;
   }

   // whats been selected?
   if (clMusic->GetCurSel() != LB_ERR)
   {
      // Process Music selection

      int Item = clMusic->GetCurSel();
      clMusic->DeleteString(Item);
      CROMMakerHelper * pRMH = m_MusModules.GetAt(m_MusModules.FindIndex(Item));
      delete pRMH;
      m_MusModules.RemoveAt(m_MusModules.FindIndex(Item));
   }
   else
   {
      // Process Sample selection
      int Item = clSamples->GetCurSel();
      clSamples->DeleteString(Item);
      CSampleEncoding * pSE = m_SampModules.GetAt(m_SampModules.FindIndex(Item));
      delete pSE;
      m_SampModules.RemoveAt(m_SampModules.FindIndex(Item));
   }

   UpdateSpaceRemaining();
}

void CROMMaker::OnSelchangeRmMusicfiles() 
{
   // Music box selected, clear Sample selections
   CListBox * cl = (CListBox *) GetDlgItem(IDC_RM_SAMPLEFILES);
   cl->SetCurSel(-1);
}

void CROMMaker::OnSelchangeRmSamplefiles() 
{
   // Sample box selected, clear Music selections
   CListBox * cl = (CListBox *) GetDlgItem(IDC_RM_MUSICFILES);
   cl->SetCurSel(-1);
}



// display how much space is left in the ROM image.
// We start with 64k less the mod player code itself and then
// subtract all of the selected music files
void CROMMaker::UpdateSpaceRemaining()
{
   int items, i;

   // update the M1 free space 64K
   int M1FreeSpace = 65536 - sizeof(Z80ModDriver);
   items = m_MusModules.GetCount();
   for (i = 0; i < items; i++)
   {
      CROMMakerHelper * pRMH = m_MusModules.GetAt(m_MusModules.FindIndex(i));
      M1FreeSpace -= pRMH->GetModuleSize();
   }
   SetDlgItemInt(IDC_RM_M1_SPACE_REMAINING, M1FreeSpace);

   // update the V1 free space 512K
   int V1FreeSpace = 512 * 1024;
   items = m_SampModules.GetCount();
   for (i = 0; i < items; i++)
   {
      CSampleEncoding * pSE = m_SampModules.GetAt(m_SampModules.FindIndex(i));
      V1FreeSpace -= pSE->GetModuleSize();
   }
   SetDlgItemInt(IDC_RM_V1_SPACE_REMAINING, V1FreeSpace);
}

// Write out the M1 ROM image for the specified music files
void CROMMaker::OnRmMakeroms() 
{
   // clear sample table
   for (int i = 0; i < SIZE_SAMPLETABLE; i++)
      m_SampleTable[i] = 0;

   m_V1ROMPos = 0;

   MakeV1();
   MakeM1();
   MakeHeaderFile();
}

void CROMMaker::MakeV1()
{
   CFile fWorkspace;
   char NullChar = 0;
   char CurDir[MAX_PATH+1];
   char Filename[MAX_PATH+1] = "202-v1";

   CFileDialog dlgFileSave(false, "BIN", NULL, OFN_HIDEREADONLY | OFN_OVERWRITEPROMPT,
      szROMFilter, NULL);
   GetCurrentDirectory(MAX_PATH, CurDir);
   dlgFileSave.m_ofn.lpstrInitialDir = CurDir;
   dlgFileSave.m_ofn.lpstrFile = Filename;
   dlgFileSave.m_ofn.nMaxFile = MAX_PATH;
   if (dlgFileSave.DoModal() == IDCANCEL)
      return;
   
   if (!fWorkspace.Open(dlgFileSave.GetPathName(), CFile::modeWrite | CFile::modeCreate))
   {
      CString ErrMsg;
      ErrMsg = "Failed to Create " + dlgFileSave.GetFileName();
      AfxMessageBox(ErrMsg);
      return;
   }

   int items = m_SampModules.GetCount();
   for (int i = 0; i < items; i++)
   {
      CSampleEncoding * pSE = m_SampModules.GetAt(m_SampModules.FindIndex(i));
      pSE->Encode(&fWorkspace);

      // get size and round up to the nearest 256byte block
      int Size = pSE->GetModuleSize();
      int RoundedSize = (Size & 0xffff00) + 0xff;
      BYTE padVal = 0x80; // ADPCM $80 = silence
      for (int pad = Size; pad <= RoundedSize; pad++)
         fWorkspace.Write(&padVal, 1);

      m_SampleTable[(i+FIRST_SAMPLE)*2] = m_V1ROMPos;
      m_V1ROMPos += (RoundedSize >> 8);
      m_SampleTable[(i+FIRST_SAMPLE)*2+1] = m_V1ROMPos;
      m_V1ROMPos ++;
   }

   // pad out to 512K here?

   // done
   fWorkspace.Close();
}

void CROMMaker::MakeM1()
{
   CFile fWorkspace;
   char NullChar = 0;
   char CurDir[MAX_PATH+1];
   char Filename[MAX_PATH+1] = "202-m1";
   int i;

   CFileDialog dlgFileSave(false, "BIN", NULL, OFN_HIDEREADONLY | OFN_OVERWRITEPROMPT,
      szROMFilter, NULL);
   GetCurrentDirectory(MAX_PATH, CurDir);
   dlgFileSave.m_ofn.lpstrInitialDir = CurDir;
   dlgFileSave.m_ofn.lpstrFile = Filename;
   dlgFileSave.m_ofn.nMaxFile = MAX_PATH;
   if (dlgFileSave.DoModal() == IDCANCEL)
      return;
   
   if (!fWorkspace.Open(dlgFileSave.GetPathName(), CFile::modeWrite | CFile::modeCreate))
   {
      CString ErrMsg;
      ErrMsg = "Failed to Create " + dlgFileSave.GetFileName();
      AfxMessageBox(ErrMsg);
      return;
   }

   // first write out the player
   fWorkspace.Write(Z80ModDriver, sizeof(Z80ModDriver));

   for (i = 0; i < SIZE_SAMPLETABLE; i ++)
   {
      WORD Val = m_SampleTable[i];
      fWorkspace.Write(&Val, 2);
   }

   // iterate though the helper objects and get each one to write its music file out
   int items = m_MusModules.GetCount();
   for (i = 0; i < items; i++)
   {
      CROMMakerHelper * pRMH = m_MusModules.GetAt(m_MusModules.FindIndex(i));
      pRMH->StoreMusic(&fWorkspace);
   }
   
   // done
   fWorkspace.Close();
}

void CROMMaker::MakeHeaderFile()
{
   CFile fWorkspace;
   char NullChar = 0;
   char CurDir[MAX_PATH+1];
   char Filename[MAX_PATH+1] = "sound.h";
   int i;

   CFileDialog dlgFileSave(false, "H", NULL, OFN_HIDEREADONLY | OFN_OVERWRITEPROMPT,
      szHeaderFileFilter, NULL);
   GetCurrentDirectory(MAX_PATH, CurDir);
   dlgFileSave.m_ofn.lpstrInitialDir = CurDir;
   dlgFileSave.m_ofn.lpstrFile = Filename;
   dlgFileSave.m_ofn.nMaxFile = MAX_PATH;
   if (dlgFileSave.DoModal() == IDCANCEL)
      return;
   
   if (!fWorkspace.Open(dlgFileSave.GetPathName(), CFile::modeWrite | CFile::modeCreate))
   {
      CString ErrMsg;
      ErrMsg = "Failed to Create " + dlgFileSave.GetFileName();
      AfxMessageBox(ErrMsg);
      return;
   }

   CString Header = ("// FM/ADPCM HEADER FILE created by MVSTracker\n\n");
   Header += "inline void send_sound_command(BYTE CommandNo)\n";
   Header += "{\n\t(*((PBYTE)0x320000)) = CommandNo;\n}\n\n";
   fWorkspace.Write(Header, Header.GetLength());

   Header = "// FM Music commands\n";
   fWorkspace.Write(Header, Header.GetLength());
   int items = m_MusModules.GetCount();
   for (i = 0; i < items; i ++)
   {
      CROMMakerHelper * pRMH = m_MusModules.GetAt(m_MusModules.FindIndex(i));
      CString ModuleName = pRMH->GetFilename();
      ModuleName.MakeUpper();
      CString Output;
      Output.Format("#define SOUND_FM_%s\t(0x%02x)\n", 
         ModuleName.Left(ModuleName.GetLength()-4), 
         0x81 + i);
      fWorkspace.Write(Output, Output.GetLength());
   }

   Header = "#define SOUND_FM_OFF\t(0xff)\n\n";
   Header+= "// ADPCM Sample commands\n";
   fWorkspace.Write(Header, Header.GetLength());
   
   items = m_SampModules.GetCount();
   for (i = 0; i < items; i ++)
   {
      CSampleEncoding * pSE = m_SampModules.GetAt(m_SampModules.FindIndex(i));
      CString ModuleName = pSE->GetFilename();
      ModuleName.MakeUpper();
      CString Output;
      Output.Format("#define SOUND_ADPCM_%s\t(0x%02x)\n", 
         ModuleName.Left(ModuleName.GetLength()-4), 
         0x04 + i);
      fWorkspace.Write(Output, Output.GetLength());
   }

   Header = "#define SOUND_ADPCM_OFF\t(0x7f)\n\n";
   fWorkspace.Write(Header, Header.GetLength());

   // done
   fWorkspace.Close();
}
