// ******************************************************************
//
// ROM Maker Helper class
//
// ******************************************************************

#include "stdafx.h"
#include "ROMMakerHelper.h"

bool CROMMakerHelper::LoadMusic(CString PathName, CString Filename)
{
   // load the file
   CFile fIn;
   int i;

   if (!fIn.Open(PathName, CFile::modeRead))
   {
      CString ErrMsg;
      ErrMsg = "Failed to Open " + Filename;
      AfxMessageBox(ErrMsg);
      return false;
   }
      
   // validate header
   char Header[sizeof(szMusicHeader)];
   fIn.Read(Header, sizeof(szMusicHeader)-1);
   if (memcmp(Header, szMusicHeader, sizeof(szMusicHeader)-1) != 0)
   {
      AfxMessageBox("Invalid MVSTracker Music file");
      return false;
   }

   // first get instruments
   for (i = 1; i < MAX_VOICES; ++i)
   {
      // get voice name and data
      CString RawInstrName = Global::ReadNullTerminatedString(&fIn);
      unsigned char Instr[INSTRLEN];

      fIn.Read(Instr, INSTRLEN);
      m_VoiceArray[i].SetInstr(Instr);
   }

   // load song: song len followed by pattern order
   fIn.Read(&m_SongLen, 1);

   memset(m_SongOrder, 0, MAX_SONG_LEN);
   BYTE SongByte;
   for (i = 0; i < m_SongLen; i ++)
   {
      fIn.Read(&SongByte, 1);
      m_SongOrder[i] = SongByte;
   }
   
   // read in number of patterns
   fIn.Read(&m_NumPatterns, 1);
   
   // now store all of the patterns
   for (i = 0; i < m_NumPatterns; i ++)
   {  
      CPattern * pPat = &m_Patterns[i];

      // get voice name and data
      CString PatName = Global::ReadNullTerminatedString(&fIn);
      unsigned char NewPat[BUILTPATSIZE];
      fIn.Read(NewPat, BUILTPATSIZE);
      pPat->SetPattern(NewPat);
   }

   fIn.Close();

   m_Filename = Filename;
   return true;
}



int CROMMakerHelper::GetModuleSize()
{
   int Size;

   // add header
   Size = 7; // "MVS0" + 2 for offset to next music + 1 for #instruments

   // add 32 byte padded instruments
   Size += (32 * GetHighestInstrNo());

   // add song info
   Size += 129; // 1 for # songlen + 128 song data

   // add patterns
   Size += (m_NumPatterns * BUILTPATSIZE);

   return Size;
}



// scans the entire song for the largest instrument number
// as this is the value we have to send to the driver to 
// save a few bytes rather than sending 32 instruments all of
// the time
int CROMMakerHelper::GetHighestInstrNo()
{
   int HighestInstrNo = 1;

   for (int Pats = 0; Pats < m_NumPatterns; Pats++)
   {
      CPattern * pPat = &m_Patterns[Pats];
      int InstrNo = pPat->GetHighestInstrNo();

      if (InstrNo > HighestInstrNo)
         HighestInstrNo = InstrNo;
   }
    
   return HighestInstrNo;
}



// Using the file stream provided write out the music file
void CROMMakerHelper::StoreMusic(CFile * fWrite)
{
   // header
   fWrite->Write("MVS0", 4);

   // write length fields
   BYTE LenHi, LenLo;
   int MusicLen = GetModuleSize() - 6;
   LenHi = MusicLen >> 8;
   LenLo = (BYTE)MusicLen;
   fWrite->Write(&LenLo, 1);
   fWrite->Write(&LenHi, 1);

   // number of instruments
   BYTE NumInstrs = (BYTE)GetHighestInstrNo();
   fWrite->Write(&NumInstrs, 1);

   for (int i = 1; i <= NumInstrs; i ++)
   {
      BYTE Instr[INSTRLEN+7]; // Instruments padded to 32 bytes
      memset(Instr, 0, INSTRLEN+7);

      m_VoiceArray[i].GetInstr(Instr);
      fWrite->Write(Instr, INSTRLEN + 7);
   }

   // song length
   fWrite->Write(&m_SongLen, 1);
   fWrite->Write(m_SongOrder, MAX_SONG_LEN);
   
   // pad the rest of the song out
   BYTE Padding[64];
   memset(Padding, 0, 64);
   fWrite->Write(Padding, 64);

   // now store all of the patterns
   for (i = 0; i < m_NumPatterns; i ++)
   {  
      CPattern * pPat = &m_Patterns[i];
      fWrite->Write(pPat->BuildPattern(false), BUILTPATSIZE);
   }
}
